# Scalar Constants.
LVL_IMPL = 115
LVL_ARCH_GOD = 114
LVL_ETERNAL_GOD = 113
LVL_GOD = 112
LVL_OVERSEER = 111
LVL_CREATOR = 110
LVL_DEMIGOD = 109
LVL_DEITY = 108
LVL_IMMORT = 107
LVL_AVATAR = 106

# debugOn()

# For CircleMUD adapters.
try: from world import LVL_IMMORT, LVL_IMPL # , LVL_GOD, LVL_GRGOD
except ImportError: pass

# Aliases.
LVL_SUPREME = LVL_IMPL
LVL_FREEZE = LVL_OVERSEER

# spells.h:
DEFAULT_STAFF_LVL   = 12
DEFAULT_WAND_LVL    = 12

CAST_UNDEFINED  = (-1)
CAST_SPELL  = 0
CAST_POTION = 1
CAST_WAND   = 2
CAST_STAFF  = 3
CAST_SCROLL = 4

MAG_DAMAGE      = (1 << 0)
MAG_AFFECTS     = (1 << 1)
MAG_UNAFFECTS   = (1 << 2)
MAG_POINTS      = (1 << 3)
MAG_ALTER_OBJS  = (1 << 4)
MAG_GROUPS      = (1 << 5)
MAG_MASSES      = (1 << 6)
MAG_AREAS       = (1 << 7)
MAG_SUMMONS     = (1 << 8)
MAG_CREATIONS   = (1 << 9)
MAG_MANUAL      = (1 << 10)


TYPE_UNDEFINED               = (-1)
SPELL_RESERVED_DBC             = 0  # SKILL NUMBER ZERO -- RESERVED

# PLAYER SPELLS -- Numbered from 1 to MAX_SPELLS

SPELL_ARMOR                   = 1 # Reserved Skill[] DO NOT CHANGE
SPELL_TELEPORT                = 2 # Reserved Skill[] DO NOT CHANGE
SPELL_BLESS                   = 3 # Reserved Skill[] DO NOT CHANGE
SPELL_BLINDNESS               = 4 # Reserved Skill[] DO NOT CHANGE
SPELL_BURNING_HANDS           = 5 # Reserved Skill[] DO NOT CHANGE
SPELL_CALL_LIGHTNING          = 6 # Reserved Skill[] DO NOT CHANGE
SPELL_CHARM                   = 7 # Reserved Skill[] DO NOT CHANGE
SPELL_CHILL_TOUCH             = 8 # Reserved Skill[] DO NOT CHANGE
SPELL_ENTANGLE                = 9 # Reserved Skill[] DO NOT CHANGE
SPELL_COLOR_SPRAY            = 10 # Reserved Skill[] DO NOT CHANGE
SPELL_CONTROL_WEATHER        = 11 # Reserved Skill[] DO NOT CHANGE
SPELL_CREATE_FOOD            = 12 # Reserved Skill[] DO NOT CHANGE
SPELL_CREATE_WATER           = 13 # Reserved Skill[] DO NOT CHANGE
SPELL_CURE_BLIND             = 14 # Reserved Skill[] DO NOT CHANGE
SPELL_CURE_CRITIC            = 15 # Reserved Skill[] DO NOT CHANGE
SPELL_CURSE                  = 17 # Reserved Skill[] DO NOT CHANGE
SPELL_DETECT_ALIGN           = 18 # Reserved Skill[] DO NOT CHANGE
SPELL_DETECT_INVIS           = 19 # Reserved Skill[] DO NOT CHANGE
SPELL_DETECT_MAGIC           = 20 # Reserved Skill[] DO NOT CHANGE
SPELL_DETECT_POISON          = 21 # Reserved Skill[] DO NOT CHANGE
SPELL_DISPEL_EVIL            = 22 # Reserved Skill[] DO NOT CHANGE
SPELL_EARTHQUAKE             = 23 # Reserved Skill[] DO NOT CHANGE
SPELL_ENCHANT_WEAPON         = 24 # Reserved Skill[] DO NOT CHANGE
SPELL_ENERGY_DRAIN           = 25 # Reserved Skill[] DO NOT CHANGE
SPELL_FIREBALL               = 26 # Reserved Skill[] DO NOT CHANGE
SPELL_HARM                   = 27 # Reserved Skill[] DO NOT CHANGE
SPELL_HEAL                   = 28 # Reserved Skill[] DO NOT CHANGE
SPELL_INVISIBLE              = 29 # Reserved Skill[] DO NOT CHANGE
SPELL_LIGHTNING_BOLT         = 30 # Reserved Skill[] DO NOT CHANGE
SPELL_LOCATE_OBJECT          = 31 # Reserved Skill[] DO NOT CHANGE
SPELL_MAGIC_MISSILE          = 32 # Reserved Skill[] DO NOT CHANGE
SPELL_POISON                 = 33 # Reserved Skill[] DO NOT CHANGE
SPELL_PROT_FROM_EVIL         = 34 # Reserved Skill[] DO NOT CHANGE
SPELL_REMOVE_CURSE           = 35 # Reserved Skill[] DO NOT CHANGE
SPELL_SANCTUARY              = 36 # Reserved Skill[] DO NOT CHANGE
SPELL_SHOCKING_GRASP         = 37 # Reserved Skill[] DO NOT CHANGE
SPELL_SLEEP                  = 38 # Reserved Skill[] DO NOT CHANGE
SPELL_STRENGTH               = 39 # Reserved Skill[] DO NOT CHANGE
SPELL_SUMMON                 = 40 # Reserved Skill[] DO NOT CHANGE
SPELL_VENTRILOQUATE          = 41 # Reserved Skill[] DO NOT CHANGE
SPELL_WORD_OF_RECALL         = 42 # Reserved Skill[] DO NOT CHANGE
SPELL_REMOVE_POISON          = 43 # Reserved Skill[] DO NOT CHANGE
SPELL_SENSE_LIFE             = 44 # Reserved Skill[] DO NOT CHANGE
SPELL_ANIMATE_CORPSE         = 45 # Reserved Skill[] DO NOT CHANGE
SPELL_DISPEL_GOOD            = 46 # Reserved Skill[] DO NOT CHANGE
SPELL_GROUP_ARMOR            = 47 # Reserved Skill[] DO NOT CHANGE
SPELL_GROUP_HEAL             = 48 # Reserved Skill[] DO NOT CHANGE
SPELL_GROUP_RECALL           = 49 # Reserved Skill[] DO NOT CHANGE
SPELL_INFRAVISION            = 50 # Reserved Skill[] DO NOT CHANGE
SPELL_WATERWALK              = 51 # Reserved Skill[] DO NOT CHANGE
SPELL_IDENTIFY               = 52
SPELL_FLY                    = 53
SPELL_FIRESHIELD             = 54
SPELL_WHUPASS                = 55
SPELL_PHASE_BLUR             = 56
SPELL_PASS_DOOR              = 57
SPELL_CREATE_SPRING          = 58
SPELL_CONTINUAL_LIGHT        = 59
SPELL_TELEPORT_OBJECT        = 60
SPELL_STONESKIN              = 61
SPELL_BARKSKIN               = 62
SPELL_ICE_STORM              = 63
SPELL_ENCHANT_ARMOR          = 64
SPELL_PORTAL                 = 65
SPELL_FORCESHIELD            = 66
SPELL_REFRESH                = 67
SPELL_ENDURANCE              = 68
SPELL_FIND_THE_PATH          = 69
SPELL_REGENERATION           = 70
SPELL_PROT_FROM_MAGIC        = 71
SPELL_EARTHMAW               = 72
SPELL_WATER_BREATHING        = 73
SPELL_FLAME_STRIKE           = 74
SPELL_CONJ_FIRE_ELEMENTAL    = 75
SPELL_CONJ_EARTH_ELEMENTAL   = 76
SPELL_CONJ_WATER_ELEMENTAL   = 77
SPELL_CONJ_AIR_ELEMENTAL     = 78
SPELL_ENERGY_BLAST           = 79
SPELL_ICE_BOLT               = 80
SPELL_RESTORATION            = 81
SPELL_IXSO_NIXSO             = 82
SPELL_REFLECT                = 83
SPELL_CHANNEL                = 84
SPELL_REGAIN_MANA            = 85
SPELL_ABSORB                 = 86
SPELL_MANA_STEAL             = 87
SPELL_GROUP_STRENGTH         = 88
SPELL_GROUP_WATER_BREATHING  = 89
SPELL_GROUP_SANCTUARY        = 90
SPELL_PEACE_SPHERE           = 91
SPELL_DIVINE_HAMMER          = 92
SPELL_POWER_LANCE            = 93
SPELL_TURN                   = 94
SPELL_CONFUSE                = 95

# Insert new spells here, up to MAX_SPELLS
MAX_SPELLS                   = 130

# PLAYER SKILLS - Numbered from MAX_SPELLS+1 to MAX_SKILLS
SKILL_BACKSTAB              = 131 # Reserved Skill[] DO NOT CHANGE
SKILL_BASH                  = 132 # Reserved Skill[] DO NOT CHANGE
SKILL_HIDE                  = 133 # Reserved Skill[] DO NOT CHANGE
SKILL_KICK                  = 134 # Reserved Skill[] DO NOT CHANGE
SKILL_PICK_LOCK             = 135 # Reserved Skill[] DO NOT CHANGE
SKILL_PUNCH                 = 136 # Reserved Skill[] DO NOT CHANGE
SKILL_RESCUE                = 137 # Reserved Skill[] DO NOT CHANGE
SKILL_SNEAK                 = 138 # Reserved Skill[] DO NOT CHANGE
SKILL_STEAL                 = 139 # Reserved Skill[] DO NOT CHANGE
SKILL_TRACK                 = 140 # Reserved Skill[] DO NOT CHANGE
SKILL_SCOUT                 = 141
SKILL_CIRCLE_AROUND         = 142
SKILL_SECOND_ATTACK         = 143
SKILL_THIRD_ATTACK          = 144
SKILL_FOURTH_ATTACK         = 145
SKILL_DUAL_WIELD            = 146
SKILL_PEEK                  = 147
SKILL_HUNT                  = 148
SKILL_DODGE                 = 149
SKILL_TUMBLE                = 150
SKILL_CRITICAL_HIT          = 151
SKILL_DANGER_SENSE          = 152
SKILL_HAND_TO_HAND          = 153
SKILL_BERSERK               = 154
SKILL_DISARM                = 155
SKILL_ELBOW                 = 156
SKILL_CHIVALRY              = 157
SKILL_DIRT_KICK             = 158
SKILL_DUAL_BACKSTAB         = 159
SKILL_MARTIAL_ARTS          = 160
SKILL_MEDITATE              = 161
SKILL_SECOND_SPELL_ATTACK   = 162
SKILL_THIRD_SPELL_ATTACK    = 163
SKILL_KNEE                  = 164
SKILL_GROIN_KICK            = 165
SKILL_LUNGE_PUNCH           = 166
SKILL_NERVE_PUNCH           = 167
SKILL_CAPTURE               = 168
SKILL_COUNTER               = 169
SKILL_MOUNTAINEERING        = 170
SKILL_DESERT_SURVIVAL       = 171
SKILL_ARCTIC_SURVIVAL       = 172
SKILL_STUN                  = 173
SKILL_FISH                  = 174

# New skills may be added here up to MAX_SKILLS (200)


##  NON-PLAYER AND OBJECT SPELLS AND SKILLS
##  The practice levels for the spells and skills below are _not_ recorded
##  in the playerfile; therefore, the intended use is for spells and skills
##  associated with objects (such as SPELL_IDENTIFY used with scrolls of
##  identify) or non-players (such as NPC-only spells).

SPELL_FIRE_BREATH            = 202
SPELL_GAS_BREATH             = 203
SPELL_FROST_BREATH           = 204
SPELL_ACID_BREATH            = 205
SPELL_LIGHTNING_BREATH       = 206

TOP_SPELL_DEFINE             = 299
# NEW NPC/OBJECT SPELLS can be inserted here up to 299

# weapon attack types
TYPE_HIT                     = 300
TYPE_STING                   = 301
TYPE_WHIP                    = 302
TYPE_SLASH                   = 303
TYPE_BITE                    = 304
TYPE_BLUDGEON                = 305
TYPE_CRUSH                   = 306
TYPE_POUND                   = 307
TYPE_CLAW                    = 308
TYPE_MAUL                    = 309
TYPE_THRASH                  = 310
TYPE_PIERCE                  = 311
TYPE_BLAST                   = 312
TYPE_PUNCH                   = 313
TYPE_STAB                    = 314
TYPE_PECK                    = 315

# new attack types can be added here - up to TYPE_SUFFERING
TYPE_CHOKING                 = 398
TYPE_SUFFERING               = 399


# saving throws
SAVING_PARA   = 0
SAVING_ROD    = 1
SAVING_PETRI  = 2
SAVING_BREATH = 3
SAVING_SPELL  = 4


# spell object types
SPELL_TYPE_SPELL   = 0
SPELL_TYPE_POTION  = 1
SPELL_TYPE_WAND    = 2
SPELL_TYPE_STAFF   = 3
SPELL_TYPE_SCROLL  = 4

# zone::reset_mode
ZONE_RESET_NEVER  = 0 # Actually, reset once on boot.
ZONE_RESET_ALWAYS = 1
ZONE_RESET_EMPTY  = 2

# zone-command::door-state::arg3
DOOR_STATE_OPEN   = 0
DOOR_STATE_CLOSED = 1
DOOR_STATE_LOCKED = 2

# Sex
SEX_NEUTRAL                    = 0                   
SEX_MALE                       = 1                   
SEX_FEMALE                     = 2                   

# Positions
POS_DEAD                       = 0                   # dead
POS_MORTALLYW                  = 1                   # mortally wounded
POS_INCAP                      = 2                   # incapacitated
POS_STUNNED                    = 3                   # stunned
POS_SLEEPING                   = 4                   # sleeping
POS_MEDITATING                 = 5                   # meditating
POS_RESTING                    = 6                   # resting
POS_SITTING                    = 7                   # sitting
POS_FIGHTING                   = 8                   # fighting
POS_STANDING                   = 9                   # standing
POS_FLYING                     = 10                  # flying (will work on this later)

# Player Flags
PLR_KILLER                     = (1 << 0)            # Player is a player-killer
PLR_THIEF                      = (1 << 1)            # Player is a player-thief
PLR_FROZEN                     = (1 << 2)            # Player is frozen
PLR_DONTSET                    = (1 << 3)            # Don't EVER set (ISNPC bit)
PLR_WRITING                    = (1 << 4)            # Player writing (board/mail/olc)
PLR_MAILING                    = (1 << 5)            # Player is writing mail
PLR_CRASH                      = (1 << 6)            # Player needs to be crash-saved
PLR_SITEOK                     = (1 << 7)            # Player has been site-cleared
PLR_NOSHOUT                    = (1 << 8)            # Player not allowed to shout/goss
PLR_NOTITLE                    = (1 << 9)            # Player not allowed to set title
PLR_DELETED                    = (1 << 10)           # Player deleted - space reusable
PLR_LOADROOM                   = (1 << 11)           # Player uses nonstandard loadroom
PLR_NOWIZLIST                  = (1 << 12)           # Player shouldn't be on wizlist
PLR_NODELETE                   = (1 << 13)           # Player shouldn't be deleted
PLR_INVSTART                   = (1 << 14)           # Player should enter game wizinvis
PLR_CRYO                       = (1 << 15)           # Player is cryo-saved (purge prog)
PLR_IT                         = (1 << 16)           # Player is it (tag)
PLR_QUESTOR                    = (1 << 17)           # Player is a questor
PLR_CREATING                   = (1 << 18)           # player is using olc
PLR_NOCLANINFO                 = (1 << 19)           # Player shouldn't be on claninfo
PLR_FULL_OLC                   = (1 << 20)           # Player has full olc access
PLR_INWAR                      = (1 << 21)           # Player is in the Arena
PLR_NOIDLE                     = (1 << 22)           # Player never force-extracted
PLR_MOBKILL                    = (1 << 23)           # Immo can kill mobs
PLR_PKOK                       = (1 << 24)           # Player partakes in PK!
PLR_NO_INFO_MSG                = (1 << 25)           # Info message shutoff
PLR_SCARRED                    = (1 << 26)           # Player has been scarred fighting
PLR_NOTDEADYET                 = (1 << 27)           # (R) Player being extracted.
PLR_AUTOENTER                  = (1 << 28)           # Player autoenters game after auth

# Mobile Flags
MOB_SPEC                       = (1 << 0)            # Mob has a callable spec-proc
MOB_SENTINEL                   = (1 << 1)            # Mob should not move
MOB_SCAVENGER                  = (1 << 2)            # Mob picks up stuff on the ground
MOB_ISNPC                      = (1 << 3)            # (R) Automatically set on all Mobs
MOB_AWARE                      = (1 << 4)            # Mob can't be backstabbed
MOB_AGGRESSIVE                 = (1 << 5)            # Mob auto-attacks everybody nearby
MOB_STAY_ZONE                  = (1 << 6)            # Mob shouldn't wander out of zone
MOB_WIMPY                      = (1 << 7)            # Mob flees if severely injured
MOB_AGGR_EVIL                  = (1 << 8)            # Auto-attack any evil PC's
MOB_AGGR_GOOD                  = (1 << 9)            # Auto-attack any good PC's
MOB_AGGR_NEUTRAL               = (1 << 10)           # Auto-attack any neutral PC's
MOB_MEMORY                     = (1 << 11)           # remember attackers if attacked
MOB_HELPER                     = (1 << 12)           # attack PCs fighting other NPCs
MOB_NOCHARM                    = (1 << 13)           # Mob can't be charmed
MOB_NOSUMMON                   = (1 << 14)           # Mob can't be summoned
MOB_NOSLEEP                    = (1 << 15)           # Mob can't be slept
MOB_NOBASH                     = (1 << 16)           # Mob can't be bashed (e.g. trees)
MOB_NOBLIND                    = (1 << 17)           # Mob can't be blinded
MOB_HUNTER                     = (1 << 18)           # Mob hunts people
MOB_NO_KILL                    = (1 << 19)           # Mob can't be attacked
MOB_NOFIRE                     = (1 << 20)           # Mob can't be damaged by fire
MOB_NOSUBDUE                   = (1 << 21)           # Mob won't give in to subduing
MOB_ON_MOB                     = (1 << 22)           # A little Mob on Mob violence
MOB_MOUNT                      = (1 << 23)           # Player can ride mob
MOB_GEN_SPEC                   = (1 << 24)           # mob uses the gen_spec routine
MOB_SILENT                     = (1 << 25)           # Mob is silent
MOB_IGNSUM                     = (1 << 26)           # Mob ignores summoned mobs
MOB_NOCORPSE                   = (1 << 27)           # Mob is animated
MOB_NOUPDATE                   = (1 << 28)           # point_update doesn't touch mob
MOB_DELETED                    = (1 << 29)           # Mob won't be written next save
MOB_NOTDEADYET                 = (1 << 30)           # (R) Mob being extracted.

# Display Flags
DISP_HIT                       = (1 << 0)            # Display current hp
DISP_MANA                      = (1 << 1)            # Display current mana
DISP_MOVE                      = (1 << 2)            # Display current move
DISP_MAXHIT                    = (1 << 3)            # Display max hit
DISP_MAXMANA                   = (1 << 4)            # Display max mana
DISP_MAXMOVE                   = (1 << 5)            # Display max move
DISP_ALIGN                     = (1 << 6)            # Display alignment meter
DISP_COND                      = (1 << 7)            # Display condition of fighting(ch)
DISP_VTBAR                     = (1 << 8)            # Display the vtbar
DISP_EXP                       = (1 << 9)            # Display exp in prompt
DISP_MOUNT                     = (1 << 10)           # Display mount
DISP_TIME                      = (1 << 11)           # Display time

# Preference Flags
PRF_BRIEF                      = (1 << 0)            # Room descs won't normally be shown
PRF_COMPACT                    = (1 << 1)            # No extra CRLF pair before prompts
PRF_DEAF                       = (1 << 2)            # Can't hear shouts
PRF_NOTELL                     = (1 << 3)            # Can't receive tells
PRF_CLANLOG                    = (1 << 4)            # Player listening to clan log?
PRF_INTINVIS                   = (1 << 5)            # Invisible to interwho
PRF_NOINTERTELL                = (1 << 6)            # Deaf to intertells
PRF_AUTOEXIT                   = (1 << 7)            # Display exits in a room
PRF_NOHASSLE                   = (1 << 8)            # Aggr mobs won't attack
PRF_QUEST                      = (1 << 9)            # On quest
PRF_SUMMONABLE                 = (1 << 10)           # Can be summoned
PRF_NOREPEAT                   = (1 << 11)           # No repetition of comm commands
PRF_HOLYLIGHT                  = (1 << 12)           # Can see in dark
PRF_COLOR                      = (1 << 13)           # Color on/off
PRF_SPEEDWALK                  = (1 << 14)           # Player using speedwalk?
PRF_NOWIZ                      = (1 << 15)           # Can't hear wizline
PRF_LOG1                       = (1 << 16)           # On-line System Log (low bit)
PRF_LOG2                       = (1 << 17)           # On-line System Log (high bit)
PRF_NOAUCT                     = (1 << 18)           # Can't hear auction channel
PRF_NOFLAME                    = (1 << 19)           # Can't hear flame channel
PRF_NOGRATZ                    = (1 << 20)           # Can't hear grats channel
PRF_ROOMFLAGS                  = (1 << 21)           # Can see room flags (ROOM_x)
PRF_NOCHAT                     = (1 << 22)           # Can't hear chat channel
PRF_NOMUSIC                    = (1 << 23)           # Can't hear music channel
PRF_AUTOLOOT                   = (1 << 24)           # Automatic loot
PRF_NOWEATHER                  = (1 << 25)           # Weather info on or off?
PRF_NOCLANTELL                 = (1 << 26)           # Can't hear clan channel
PRF_ALL_CLANS                  = (1 << 27)           # Can hear all clan channels
PRF_GAG                        = (1 << 28)           # Missed fighting messages don't show
PRF_NOARENA                    = (1 << 29)           # Player is deaf to the arena channel
PRF_LOG3                       = (1 << 30)           # On-line System log, debuggory
PRF_STAYPAGE                   = (1 << 31)           # Do not auto-quit last string page

# Preferences Flags 2
PRF2_NOCLANTITLE               = (1 << 0)            # Player's clantitle will not show
PRF2_CLANRECALL                = (1 << 1)            # Player recalls to hometown or clan?
PRF2_AFK                       = (1 << 2)            # Player is afk
PRF2_NOPROMPT                  = (1 << 3)            # Prompt will not be redrawn
PRF2_MYTITLE                   = (1 << 4)            # Title doesn't changed through lvls
PRF2_AUTOASSIST                = (1 << 5)            # Automatically assist fighting chs
PRF2_SUBDUE                    = (1 << 6)            # Character subdues in fights
PRF2_AUTOSAC                   = (1 << 7)            # Character automatically sacrifices
PRF2_NOANNOUNCE                = (1 << 8)            # Character doesn't receive info
PRF2_COOPERATE                 = (1 << 9)            # Character cooperates with others
PRF2_NOMOBLSTTELL              = (1 << 10)           # Character has mob tells in last
PRF2_AUTOGOLD                  = (1 << 11)           # Character automatically loots gold
PRF2_NOBLINK                   = (1 << 12)           # Character ignores blinking text?
PRF2_SMALL_MENUS               = (1 << 13)           # Player receives small or large?
PRF2_TIME_CMDS                 = (1 << 14)           # Player receives duration message
PRF2_WIZDIAGNOSE               = (1 << 15)           # Wizard sees verbose fight messgs
PRF2_SNUFF                     = (1 << 16)           # Hehe, used in conjunction with gag
PRF2_HINTS                     = (1 << 17)           # See random hints given by mud

# Affect Bits
AFF_BLIND                      = (1 << 0)            # (R) Char is blind
AFF_INVISIBLE                  = (1 << 1)            # Char is invisible
AFF_DETECT_ALIGN               = (1 << 2)            # Char is sensitive to align
AFF_DETECT_INVIS               = (1 << 3)            # Char can see invis chars
AFF_DETECT_MAGIC               = (1 << 4)            # Char is sensitive to magic
AFF_SENSE_LIFE                 = (1 << 5)            # Char can sense hidden life
AFF_WATERWALK                  = (1 << 6)            # Char can walk on water
AFF_SANCTUARY                  = (1 << 7)            # Char protected by sanct.
AFF_GROUP                      = (1 << 8)            # (R) Char is grouped
AFF_CURSE                      = (1 << 9)            # Char is cursed
AFF_INFRAVISION                = (1 << 10)           # Char can see in dark
AFF_POISON                     = (1 << 11)           # (R) Char is poisoned
AFF_PROTECT_EVIL               = (1 << 12)           # Char protected from evil
AFF_PROTECT_GOOD               = (1 << 13)           # Char protected from good
AFF_SLEEP                      = (1 << 14)           # (R) Char magically asleep
AFF_NOTRACK                    = (1 << 15)           # Char can't be tracked
AFF_FLY                        = (1 << 16)           # Character is flying
AFF_FIRESHIELD                 = (1 << 17)           # Char aff'd by fireshield
AFF_SNEAK                      = (1 << 18)           # Char can move quietly
AFF_HIDE                       = (1 << 19)           # Char is hidden
AFF_PHASE                      = (1 << 20)           # Char is affected by phase
AFF_CHARM                      = (1 << 21)           # Char is charmed
AFF_TRANSLUCENT                = (1 << 22)           # Char walks through doors
AFF_SECOND_ATTACK              = (1 << 23)           # mobile has second attack
AFF_THIRD_ATTACK               = (1 << 24)           # mobile has third attack
AFF_FOURTH_ATTACK              = (1 << 25)           # mobile has fourth attack
AFF_MEDITATING                 = (1 << 26)           # mobile is meditating
AFF_HIT_BY_LIGHTNING           = (1 << 27)           # between standard wait
AFF_REFLECT                    = (1 << 28)           # reflects most spells
AFF_ABSORB                     = (1 << 29)           # absorb as move/mn/hit
AFF_ENTANGLE                   = (1 << 30)           # mobile cannot flee

# NOTE: Don't confuse these constants with the ITEM_ bitvectors
#   which control the valid places you can wear a piece of equipment
WEAR_LIGHT              = 0
WEAR_FINGER_R           = 1
WEAR_FINGER_L           = 2
WEAR_NECK_1             = 3
WEAR_NECK_2             = 4
WEAR_BODY               = 5
WEAR_HEAD               = 6
WEAR_LEGS               = 7
WEAR_FEET               = 8
WEAR_HANDS              = 9
WEAR_ARMS               = 10
WEAR_SHIELD             = 11
WEAR_ABOUT              = 12
WEAR_WAIST              = 13
WEAR_WRIST_R            = 14
WEAR_WRIST_L            = 15
WEAR_WIELD_1            = 16
WEAR_HOLD               = 17
WEAR_EAR_R              = 18
WEAR_EAR_L              = 19
WEAR_FACE               = 20
WEAR_WIELD_2            = 21
WEAR_WIELD_BOTH         = 22
WEAR_RESTRAINT          = 23

NUM_WEARS               = 24       # This must be the # of eq positions!!

# room-related defines

# The cardinal directions: used as index to room_data.dir_option[]
NORTH                          = 0                   
EAST                           = 1                   
SOUTH                          = 2                   
WEST                           = 3                   
UP                             = 4                   
DOWN                           = 5                   


# Zone flags: used in zone_data.zone_flags
ZONE_OLC                       = (1 << 0)            # OLC zone
ZONE_GOD                       = (1 << 1)            # A God-only zone
ZONE_CLAN                      = (1 << 2)            # Clan zone
ZONE_REMORT                    = (1 << 3)            # RemortOnly zone

# Room flags: used in room_data.room_flags
# WARNING: In the world files, NEVER set the bits marked "R" ("Reserved")
ROOM_DARK                      = (1 << 0)            # Dark
ROOM_DEATH                     = (1 << 1)            # Death trap
ROOM_NOMOB                     = (1 << 2)            # MOBs not allowed
ROOM_INDOORS                   = (1 << 3)            # Indoors
ROOM_PEACEFUL                  = (1 << 4)            # Violence not allowed
ROOM_SOUNDPROOF                = (1 << 5)            # Shouts, gossip blocked
ROOM_NOTRACK                   = (1 << 6)            # Track won't go through
ROOM_NOMAGIC                   = (1 << 7)            # Magic not allowed
ROOM_TUNNEL                    = (1 << 8)            # room for only 1 pers
ROOM_PRIVATE                   = (1 << 9)            # Can't teleport in
ROOM_GODROOM                   = (1 << 10)           # LVL_GOD+ only allowed
ROOM_HOUSE                     = (1 << 11)           # (R) Room is a house
ROOM_HOUSE_CRASH               = (1 << 12)           # (R) House needs saving
ROOM_ATRIUM                    = (1 << 13)           # (R) The door to a house
ROOM_OLC                       = (1 << 14)           # (R) Modifyable/!compress
ROOM_BFS_MARK                  = (1 << 15)           # (R) breath-first srch mrk
ROOM_IMP_ROOM                  = (1 << 16)           # LVL_LSIMP+ only allowed
ROOM_ARENA                     = (1 << 17)           # Player killing allowed
ROOM_NO_TELEPORT               = (1 << 18)           # Can't teleport in
ROOM_NO_QUIT                   = (1 << 19)           # sent back to start room
ROOM_CLAN                      = (1 << 20)           # clan entrance
ROOM_REGEN                     = (1 << 21)           # Regen room
ROOM_NOMORT                    = (1 << 22)           # mortals can't enter
ROOM_NORECALL                  = (1 << 23)           # mortals can't recall from
ROOM_DELETED                   = (1 << 24)           # Room won't be saved


# Exit info: used in room_data.dir_option.exit_info
EX_ISDOOR                      = (1 << 0)            # Exit is a door
EX_CLOSED                      = (1 << 1)            # The door is closed
EX_LOCKED                      = (1 << 2)            # The door is locked
EX_PICKPROOF                   = (1 << 3)            # Lock can't be picked
EX_HIDDEN                      = (1 << 4)            # Secret door
EX_MAG_RESIST                  = (1 << 5)            # Magic resistant door


# Sector types: used in room_data.sector_type
SECT_INSIDE                    = 0                   # Indoors
SECT_CITY                      = 1                   # In a city
SECT_FIELD                     = 2                   # In a field
SECT_FOREST                    = 3                   # In a forest
SECT_HILLS                     = 4                   # In the hills
SECT_MOUNTAIN                  = 5                   # On a mountain
SECT_WATER_SWIM                = 6                   # Swimmable water
SECT_WATER_NOSWIM              = 7                   # Water - need a boat
SECT_UNDERWATER                = 8                   # Underwater
SECT_FLYING                    = 9                   # Mid Air
SECT_DESERT                    = 10                  # Desert
SECT_JUNGLE                    = 11                  # Jungle
SECT_SNOW                      = 12                  # Snow covered
SECT_SWAMP                     = 13                  # Swamp


# object-related defines

# Item types: used by obj_data.obj_flags.type_flag
ITEM_LIGHT                     = 1                   # Item is a light source
ITEM_SCROLL                    = 2                   # Item is a scroll
ITEM_WAND                      = 3                   # Item is a wand
ITEM_STAFF                     = 4                   # Item is a staff
ITEM_WEAPON                    = 5                   # Item is a weapon
ITEM_FIREWEAPON                = 6                   # Unimplemented
ITEM_MISSILE                   = 7                   # Unimplemented
ITEM_TREASURE                  = 8                   # Item is a treasure, not gold
ITEM_ARMOR                     = 9                   # Item is armor
ITEM_POTION                    = 10                  # Item is a potion
ITEM_WORN                      = 11                  # Unimplemented
ITEM_OTHER                     = 12                  # Misc object
ITEM_TRASH                     = 13                  # Trash - shopkeeps won't buy
ITEM_TRAP                      = 14                  # Unimplemented
ITEM_CONTAINER                 = 15                  # Item is a container
ITEM_NOTE                      = 16                  # Item is note
ITEM_DRINKCON                  = 17                  # Item is a drink container
ITEM_KEY                       = 18                  # Item is a key
ITEM_FOOD                      = 19                  # Item is food
ITEM_MONEY                     = 20                  # Item is money (gold)
ITEM_PEN                       = 21                  # Item is a pen
ITEM_BOAT                      = 22                  # Item is a boat
ITEM_FOUNTAIN                  = 23                  # Item is a fountain
ITEM_PORTAL                    = 24                  # Item is a portal
ITEM_PILL                      = 25                  # Item is a pill


# Take/Wear flags: used by obj_data.obj_flags.wear_flags
ITEM_WEAR_TAKE                 = (1 << 0)            # Item can be takes
ITEM_WEAR_FINGER               = (1 << 1)            # Can be worn on finger
ITEM_WEAR_NECK                 = (1 << 2)            # Can be worn around neck
ITEM_WEAR_BODY                 = (1 << 3)            # Can be worn on body
ITEM_WEAR_HEAD                 = (1 << 4)            # Can be worn on head
ITEM_WEAR_LEGS                 = (1 << 5)            # Can be worn on legs
ITEM_WEAR_FEET                 = (1 << 6)            # Can be worn on feet
ITEM_WEAR_HANDS                = (1 << 7)            # Can be worn on hands
ITEM_WEAR_ARMS                 = (1 << 8)            # Can be worn on arms
ITEM_WEAR_SHIELD               = (1 << 9)            # Can be used as a shield
ITEM_WEAR_ABOUT                = (1 << 10)           # Can be worn about body
ITEM_WEAR_WAIST                = (1 << 11)           # Can be worn around waist
ITEM_WEAR_WRIST                = (1 << 12)           # Can be worn on wrist
ITEM_WEAR_WIELD                = (1 << 13)           # Can be wielded
ITEM_WEAR_HOLD                 = (1 << 14)           # Can be held
ITEM_WEAR_EAR                  = (1 << 15)           # Can be worn in ear
ITEM_WEAR_FACE                 = (1 << 16)           # Can be worn on face
ITEM_WEAR_RESTRAINT            = (1 << 17)           # Restraint (captured chars)

# Anticlass object flags: used by obj_data.obj_flags.anti_flags
ANTI_WIZARD                    = (1 << 0)            # Wizards cannot use
ANTI_CLERIC                    = (1 << 1)            # Clerics cannot use
ANTI_ROGUE                     = (1 << 2)            # Rogues cannot use
ANTI_KNIGHT                    = (1 << 3)            # Knights cannot use
ANTI_RANGER                    = (1 << 4)            # Rangers cannot use
ANTI_PALADIN                   = (1 << 5)            # Paladins cannot use
ANTI_MONK                      = (1 << 6)            # Monks cannot use
ANTI_NINJA                     = (1 << 7)            # Ninjas cannot use


# Extra object flags: used by obj_data.obj_flags.extra_flags
ITEM_GLOW                      = (1 << 0)            # Item is glowing
ITEM_HUM                       = (1 << 1)            # Item is humming
ITEM_NORENT                    = (1 << 2)            # Item cannot be rented
ITEM_NODONATE                  = (1 << 3)            # Item cannot be donated
ITEM_NOINVIS                   = (1 << 4)            # Item cannot be made invis
ITEM_INVISIBLE                 = (1 << 5)            # Item is invisible
ITEM_MAGIC                     = (1 << 6)            # Item is magical
ITEM_NODROP                    = (1 << 7)            # Item is cursed: can't drop
ITEM_BLESS                     = (1 << 8)            # Item is blessed
ITEM_ANTI_GOOD                 = (1 << 9)            # Not usable by good people
ITEM_ANTI_EVIL                 = (1 << 10)           # Not usable by evil people
ITEM_ANTI_NEUTRAL              = (1 << 11)           # Not usable by neutral people
ITEM_TWO_HANDED                = (1 << 12)           # Weapon requires both hands
ITEM_MAG_CREATED               = (1 << 13)           # Created from magic
ITEM_FLAMING                   = (1 << 14)           # Weapon is flaming
ITEM_POISONED                  = (1 << 15)           # Weapon is poisoned
ITEM_NOSELL                    = (1 << 16)           # Shopkeepers won't touch it
ITEM_NO_MOB_TAKE               = (1 << 17)           # Scavenger mobs won't touch
ITEM_QUEST_EQ                  = (1 << 18)           # Item is quest equipment
ITEM_NODECAY                   = (1 << 19)           # Item Will Never Decay
ITEM_NO_LOCATE                 = (1 << 20)           # Can't be located by spells
ITEM_DONATED                   = (1 << 21)           # Object was donated
ITEM_NOAUCTION                 = (1 << 22)           # Object cannot be auctioned
ITEM_SHINE                     = (1 << 23)           # Object glows as light source
ITEM_UNAPPROVED                = (1 << 24)           # Objects from OLC zones that have not been approved yet
ITEM_HIDDEN                    = (1 << 25)           # Can't see it in room
ITEM_NO_DISARM                 = (1 << 26)           # Item cannot be disarms
ITEM_TRANSPARENT               = (1 << 27)           # Item can be seen through
ITEM_BIODEGRADABLE             = (1 << 28)           # Item won't biodegrade
ITEM_DELETED                   = (1 << 29)           # Item won't be written
ITEM_NOIDENTIFY                = (1 << 30)           # Item can't be identified
ITEM_DECAYING                  = (1 << 31)           # Item Is Decaying


# Defines for restraining objects
RESTRAIN_COMM                  = 1                   
RESTRAIN_POS                   = 2                   
RESTRAIN_MOVE                  = 3                   
RESTRAIN_FIGHT                 = 4                   
RESTRAIN_SKILL                 = 5                   

R_COMM_BIT                     = (1 << 0)            
R_POS_BIT                      = (1 << 1)            
R_MOVE_BIT                     = (1 << 2)            
R_FIGHT_BIT                    = (1 << 3)            
R_SKILL_BIT                    = (1 << 4)            


# Modifier constants used with obj affects ('A' fields)
APPLY_NONE                     = 0                   # No effect
APPLY_STR                      = 1                   # Apply to strength
APPLY_DEX                      = 2                   # Apply to dexterity
APPLY_INT                      = 3                   # Apply to intelligence
APPLY_WIS                      = 4                   # Apply to wisdom
APPLY_CON                      = 5                   # Apply to constitution
APPLY_CHA                      = 6                   # Apply to charisma
APPLY_CLASS                    = 7                   # Reserved
APPLY_LEVEL                    = 8                   # Reserved
APPLY_AGE                      = 9                   # Apply to age
APPLY_CHAR_WEIGHT              = 10                  # Apply to weight
APPLY_CHAR_HEIGHT              = 11                  # Apply to height
APPLY_MANA                     = 12                  # Apply to max mana
APPLY_HIT                      = 13                  # Apply to max hit points
APPLY_MOVE                     = 14                  # Apply to max move points
APPLY_GOLD                     = 15                  # Reserved
APPLY_EXP                      = 16                  # Reserved
APPLY_AC                       = 17                  # Apply to Armor Class
APPLY_HITROLL                  = 18                  # Apply to hitroll
APPLY_DAMROLL                  = 19                  # Apply to damage roll
APPLY_SAVING_PARA              = 20                  # Apply to save throw: paralz
APPLY_SAVING_ROD               = 21                  # Apply to save throw: rods
APPLY_SAVING_PETRI             = 22                  # Apply to save throw: petrif
APPLY_SAVING_BREATH            = 23                  # Apply to save throw: breath
APPLY_SAVING_SPELL             = 24                  # Apply to save throw: spells

# Modifier constants used for rooms
ROOM_APPLY_NONE                = 0                   # No efefct

# Container flags - value[1]
CONT_CLOSEABLE                 = (1 << 0)            # Container can be closed
CONT_PICKPROOF                 = (1 << 1)            # Container is pickproof
CONT_CLOSED                    = (1 << 2)            # Container is closed
CONT_LOCKED                    = (1 << 3)            # Container is locked

# Special container value[3]
CONT_NORMAL                    = 0                   # Not a special container
CONT_CORPSE                    = 1                   # Container is a corpse
CONT_PILE                      = 2                   # Container is a pile
CONT_VOMIT                     = 3                   # Container is vomit

# Some different kind of liquids for use in values of drink containers
LIQ_WATER                      = 0                   
LIQ_BEER                       = 1                   
LIQ_WINE                       = 2                   
LIQ_ALE                        = 3                   
LIQ_DARKALE                    = 4                   
LIQ_WHISKY                     = 5                   
LIQ_LEMONADE                   = 6                   
LIQ_FIREBRT                    = 7                   
LIQ_LOCALSPC                   = 8                   
LIQ_SLIME                      = 9                   
LIQ_MILK                       = 10                  
LIQ_TEA                        = 11                  
LIQ_COFFE                      = 12                  
LIQ_BLOOD                      = 13                  
LIQ_SALTWATER                  = 14                  
LIQ_CLEARWATER                 = 15                  
LIQ_CHAMPAGNE                  = 16                  


# other miscellaneous defines


# Player conditions
DRUNK                          = 0                   
FULL                           = 1                   
THIRST                         = 2                   


# Sun state for weather_data
SUN_DARK                       = 0                   
SUN_RISE                       = 1                   
SUN_LIGHT                      = 2                   
SUN_SET                        = 3                   


# Sky conditions for weather_data
SKY_CLOUDLESS                  = 0                   
SKY_CLOUDY                     = 1                   
SKY_RAINING                    = 2                   
SKY_LIGHTNING                  = 3                   


# Rent codes
RENT_UNDEF                     = 0                   
RENT_CRASH                     = 1                   
RENT_RENTED                    = 2                   
RENT_CRYO                      = 3                   
RENT_FORCED                    = 4                   
RENT_TIMEDOUT                  = 5                   

# Special vnums
MOB_ANIMATED_CORPSE            = 7                   


# Maps, Tables & Vectors.

# imm_levels
_level_names = {LVL_IMPL        : 'Implementor',
                LVL_ARCH_GOD    : 'Arch God'   ,
                LVL_ETERNAL_GOD : 'Eternal God',
                LVL_GOD         : 'God'        ,
                LVL_OVERSEER    : 'Overseer'   ,
                LVL_CREATOR     : 'Creator'    ,
                LVL_DEMIGOD     : 'Demi-God'   ,
                LVL_DEITY       : 'Deity'      ,
                LVL_IMMORT      : 'Immortal'   ,
                LVL_AVATAR      : 'Avatar'}

# attack_hit_type = {singular, plural}
_attack_hit_text = [('hit'      , 'hits'     ),
                    ('sting'    , 'stings'   ),
                    ('whip'     , 'whips'    ),
                    ('slash'    , 'slashes'  ),
                    ('bite'     , 'bites'    ),
                    ('bludgeon' , 'bludgeons'),
                    ('crush'    , 'crushes'  ),
                    ('pound'    , 'pounds'   ),
                    ('claw'     , 'claws'    ),
                    ('maul'     , 'mauls'    ),
                    ('thrash'   , 'thrashes' ),
                    ('pierce'   , 'pierces'  ),
                    ('blast'    , 'blasts'   ),
                    ('punch'    , 'punches'  ),
                    ('stab'     , 'stabs'    ),
                    ('peck'     , 'pecks'    )]

_dam_weapons = [dict(to_room   = '',
                     to_char   = '',
                     to_victim = '')]

##    {
##      "$n tries to #w $N, but misses.",
##      "You try to #w $N, but miss.",
##      "$n tries to #w you, but misses."
##    },
##
##    {
##      "$n barely #W $N.",
##      "You barely #w $N.",
##      "$n barely #W you."
##    },
##
##    {
##      "$n scratches $N with $s #w.",
##      "You scratch $N as you #w $M.",
##      "$n scratches you as $e #W you."
##    },
##
##    {
##      "$n #W $N.",
##      "You #w $N.",
##      "$n #W you."
##    },
##
##    {
##      "$n #W $N hard.",
##      "You #w $N hard.",
##      "$n #W you hard."
##    },
##
##    {
##      "$n #W $N very hard.",
##      "You #w $N very hard.",
##      "$n #W you very hard."
##    },
##
##    {
##      "$n #W $N extremely hard.",
##      "You #w $N extremely hard.",
##      "$n #W you extremely hard."
##    },
##
##    {
##      "$n massacres $N to small fragments with $s #w.",
##      "You massacre $N to small fragments with your #w.",
##      "$n massacres you to small fragments with $s #w."
##    },
##
##    {
##      "$n maims $N with $s #w!",
##      "You maim $N with your #w!",
##      "$n maims you with $s #w!"
##    },
##
##    {
##      "$n MUTILATES $N with $s harsh #w!",
##      "You MUTILATE $N with your harsh #w!",
##      "$n MUTILATES you with $s harsh #w!"
##    },
##
##    {
##      "$n OBLITERATES $N with $s deadly #w!",
##      "You OBLITERATE $N with your deadly #w!",
##      "$n OBLITERATES you with $s deadly #w!"
##    },
##
##    {
##      "$n ANNIHILATES $N with $s wicked #w!",
##      "You ANNIHILATE $N with your wicked #w!",
##      "$n ANNIHILATES you with $s wicked #w!"
##    },
##
##    {
##      "$n VAPORIZES $N with $s DEVASTATING #w!!",
##      "You VAPORIZE $N with your DEVASTATING #w!!",
##      "$n VAPORIZES you with $s DEVASTATING #w!!"
##    }

_move_list = [('walk'   ,'walks'   ),
              ('ooze'   ,'oozes'   ),
              ('stomp'  ,'stomps'  ),
              ('hop'    ,'hops'    ),
              ('slide'  ,'slides'  ),
              ('stroll' ,'strolls' ),
              ('wander' ,'wanders' ),
              ('crawl'  ,'crawls'  ),
              ('slither','slithers'),
              ('float'  ,'floats'  ),
              ('bounce' ,'bounces' ),
              ('run'    ,'run'     ),
              ('roll'   ,'rolls'   ),
              ('slink'  ,'slinks'  ),
              ('scurry' ,'scurries'),
              ('waddle' ,'waddles' ),
              ('glide'  ,'glides'  ),
              ('fly'    ,'flies'   ),
              ('swim'   ,'swims'   ),
              ('gallop' ,'gallops' )]

_movement_loss = {0  : 1,               # Inside
                  1  : 1,               # City
                  2  : 2,               # Field
                  3  : 3,               # Forest
                  4  : 4,               # Hills
                  5  : 6,               # Mountains
                  6  : 4,               # Swimming
                  6  : 1,               # Unswimable
                  7  : 2,               # Underwater
                  8  : 0,               # Flying
                  9  : 4,               # Desert
                  10 : 5,               # Jungle
                  11 : 5,               # Snow
                  12 : 5}               # Swamp

zoneFlagNamesDict = {
    1:"Olc",
    2:"Immortal",
    4:"Clan",
    8:"RemortOnly"
}

    # XXX These are in the wrong order, actually, but they're more familiar.
    ##    resetModes = {
    ##        0:"Never Reset",
    ##        1:"Reset When Empty",
    ##        2:"Always Reset"
    ##    }

_reset_mode_names = {ZONE_RESET_NEVER  : 'Once on boot',
                     ZONE_RESET_ALWAYS : 'Always'      ,
                     ZONE_RESET_EMPTY  : 'When Empty'}

_continents = {0: 'Marklar',
               1: 'Landrea'}

_directions = 'neswud'

_exit_bits = {0: "Door"     ,
              1: "Closed"   ,
              2: "Locked"   ,
              3: "PickProof",
              4: "Hidden"   ,
              5: "NoMagic"  }

_room_sectors = {
        0:"Inside",
        1:"City",
        2:"Field",
        3:"Forest",
        4:"Hills",
        5:"Mountains",
        6:"Water (Swim)",
        7:"Water (No Swim)",
        8:"Underwater",
        9:"Mid Air",
        10:"Desert",
        11:"Jungle",
        12:"Snow",
        13:"Swamp"
        }

roomFlagNamesDict = {
    1:"Dark",
    2:"DeathTrap",
    4:"NoMob",
    8:"Indoors",
    16:"Peaceful",
    32:"Soundproof",
    64:"NoTrack",
    128:"NoMagic",
    512:"Tunnel",
    1024:"Private",
    2048:"GodRoom",
    4096:"House",
    8192:"HouseCrashSave",
    16384:"Atrium",
    32768:"Olc",
    65536:"*", # BFS MARK
    131072:"ImplementorRoom",
    262144:"Arena",
    524288:"NoTeleport",
    1048576:"NoQuit",
    2097152:"ClanHouseEntrance",
    4195304:"Regeneration",
    8388608:"NoMortalPC",
    16777216:"Deleted"
    }

_door_states = {DOOR_STATE_OPEN   : 'Open'  ,
                DOOR_STATE_CLOSED : 'Closed',
                DOOR_STATE_LOCKED : 'Locked'}

_wear_positions = {WEAR_LIGHT       : 'Held as light'   ,
                   WEAR_FINGER_R    : 'On Right Finger' ,
                   WEAR_FINGER_L    : 'On Left Finger'  ,
                   WEAR_NECK_1      : 'Around Neck'     ,
                   WEAR_NECK_2      : 'Around Breast'   ,
                   WEAR_BODY        : 'On Body'         ,
                   WEAR_HEAD        : 'On Head'         ,
                   WEAR_LEGS        : 'On Legs'         ,
                   WEAR_FEET        : 'On Feet'         ,
                   WEAR_HANDS       : 'On Hands'        ,
                   WEAR_ARMS        : 'On Arms'         ,
                   WEAR_SHIELD      : 'Used as Shield'  ,
                   WEAR_ABOUT       : 'About Body'      ,
                   WEAR_WAIST       : 'About Waist'     ,
                   WEAR_WRIST_R     : 'On Right Wrist'  ,
                   WEAR_WRIST_L     : 'On Left Wrist'   ,
                   WEAR_WIELD_1     : 'Primary Wield'   ,
                   WEAR_HOLD        : 'Held'            ,
                   WEAR_EAR_R       : 'Right Ear'       ,
                   WEAR_EAR_L       : 'Left Ear'        ,
                   WEAR_FACE        : 'On Face'         ,
                   WEAR_WIELD_2     : 'Secondary Wield' ,
                   WEAR_WIELD_BOTH  : 'Two-Handed Wield',
                   WEAR_RESTRAINT   : 'Restraint'       }

_wear_bits = {0 : 'Take'  ,
              1 : 'Finger',
              2 : 'Neck'  ,
              3 : 'Body'  ,
              4 : 'Head'  ,
              5 : 'Legs'  ,
              6 : 'Feet'  ,
              7 : 'Hands' ,
              8 : 'Arms'  ,
              9 : 'Shield',
              10: 'About' ,
              11: 'Waist' ,
              12: 'Wrist' ,
              13: 'Wield' ,
              14: 'Hold'  ,
              15: 'Ear'   ,
              16: 'Face'  ,
              17: 'Restraint'}

_positions = {0 : "Dead"             ,
              1 : "Mortally wounded" ,
              2 : "Incapacitated"    ,
              3 : "Stunned"          ,
              4 : "Sleeping"         ,
              5 : "Meditating"       ,
              6 : "Resting"          ,
              7 : "Sitting"          ,
              8 : "Fighting"         ,
              9 : "Standing"         }

_display_bits = {0 : "HitPoints"    ,
                 1 : "Mana"         ,
                 2 : "Movement"     ,
                 3 : "MaxHitPoints" ,
                 4 : "MaxMana"      ,
                 5 : "MaxMovement"  ,
                 6 : "Alignment"    ,
                 7 : "AutoDiagnose" ,
                 8 : "VTBar"        ,
                 9 : "Experience"   ,
                 10: "Mount"        ,
                 11: "Time"         }

_player_bits = {0 : 'Killer'    ,
                1 : 'Thief'     ,
                2 : 'Frozen'    ,
                3 : 'DONTSET'   ,
                4 : 'Writing'   ,
                5 : 'Mailing'   ,
                6 : 'CrashSave' ,
                7 : 'SiteOk'    ,
                8 : 'Muted'     ,
                9 : 'NoTitle'   ,
                10: 'Deleted'   ,
                11: 'Loadroom'  ,
                12: 'NoWizlist' ,
                13: 'NoDelete'  ,
                14: 'InvisStart',
                15: "Cryo'd"    ,
                16: 'It'        ,
                17: 'Questor'   ,
                18: 'Creating'  ,
                19: 'NoClanInfo',
                20: 'FullOlcPerm',
                21: 'InWar'     ,
                22: 'NoIdle'    ,
                23: 'MobKill'   ,
                24: 'PKOK'      ,
                25: 'NoInfo'    ,
                26: 'Scarred'   ,
                27: 'MortalWounds',
                28: 'AutoEnter' }

_npc_bits = {0 : 'SpecProc'       ,
             1 : 'Sentinel'       ,
             2 : 'Scavenger'      ,
             3 : 'IsNPC'          ,
             4 : 'NoBackstab'     ,
             5 : 'Aggressive'     ,
             6 : 'StayInZone'     ,
             7 : 'Wimpy'          ,
             8 : 'AggressiveEvil' ,
             9 : 'AggressiveGood' ,
             10: 'AggressiveNeutral',
             11: 'Memory'         ,
             12: 'Helper'         ,
             13: 'NoCharm'        ,
             14: 'NoSummon'       ,
             15: 'NoSleep'        ,
             16: 'NoBash'         ,
             17: 'NoBlind'        ,
             18: 'Hunter'         ,
             19: 'Invulnerable'   ,
             20: 'NoFire'         ,
             21: 'NoSubdue'       ,
             22: 'MobViolence'    ,
             23: 'Mount'          ,
             24: 'GenSpec'        ,
             25: 'Silent'         ,
             26: 'IgnoreSummons'  ,
             27: 'NoCorpse'       ,
             28: 'NoUpdate'       ,
             29: 'Deleted'        }

_preference_bits = {0 : 'Brief'     ,
                    1 : 'Compact'   ,
                    2 : 'NoShout'   ,
                    3 : 'NoTell'    ,
                    4 : 'ClanLog'   ,
                    5 : 'Invisible' ,
                    6 : '!UNUSED!'  ,
                    7 : 'AutoExits' ,
                    8 : 'NoHassle'  ,
                    9 : 'Quest'     ,
                    10: 'Summon'    ,
                    11: 'NoRepeat'  ,
                    12: 'HolyLight' ,
                    13: 'AnsiColor' ,
                    14: 'Speedwalk' ,
                    15: 'NoWiznet'  ,
                    16: 'SyslogBit1',
                    17: 'SyslogBit2',
                    18: 'NoAuction' ,
                    19: 'NoFlame'   ,
                    20: 'NoGratz'   ,
                    21: 'RoomFlags' ,
                    22: 'NoChat'    ,
                    23: 'NoMusic'   ,
                    24: 'AutoLoot'  ,
                    25: 'NoWeather' ,
                    26: 'NoClanTell',
                    27: 'AllClanTells',
                    28: 'MessageGag',
                    29: 'NoArena'   ,
                    30: 'SyslogBit3',
                    31: 'StayPage'  }

_preference_bits2 = {0 : 'NoClanTitle'      ,
                     1 : 'ClanRecall'       ,
                     2 : 'AFK'              ,
                     3 : 'NoPrompt'         ,
                     4 : 'KeepMyTitle'      ,
                     5 : 'AutoAssist'       ,
                     6 : 'Subdues'          ,
                     7 : 'AutoSacrifice'    ,
                     8 : 'IgnoreAnnouncements',
                     9 : 'Cooperates'       ,
                     10: 'NoMobLastTell'    ,
                     11: 'AutoGold'         ,
                     12: 'NoBlink'          ,
                     13: 'SmallMenus'       ,
                     14: 'CommandTiming'    ,
                     15: 'MessageSnuff'     ,
                     16: 'GameHints'        }

_affected_bits = {0 : 'Blind'           ,
                  1 : 'Invisible'       ,
                  2 : 'DetectAlign'     ,
                  3 : 'DetectInvis'     ,
                  4 : 'DetectMagic'     ,
                  5 : 'SenseLife'       ,
                  6 : 'Waterwalk'       ,
                  7 : 'Sanctuary'       ,
                  8 : 'Grouped'         ,
                  9 : 'Cursed'          ,
                  10: 'Infravision'     ,
                  11: 'Poisoned'        ,
                  12: 'ProtectionFromEvil',
                  13: 'ProtectionFromGood',
                  14: 'Sleep'           ,
                  15: 'NoTrack'         ,
                  16: 'Flying'          ,
                  17: 'Fireshield'      ,
                  18: 'Sneaking'        ,
                  19: 'Hiding'          ,
                  20: 'PhaseBlur'       ,
                  21: 'Charmed'         ,
                  22: 'Etherealized'    ,
                  23: 'SecondAttack'    ,
                  24: 'ThirdAttack'     ,
                  25: 'FourthAttack'    ,
                  26: 'Meditating'      ,
                  27: 'HitByLightning'  ,
                  28: 'Reflect'         ,
                  29: 'Absorb'          ,
                  30: 'Entangled'       }

_class_names = {0: "wizard"  ,
                1: "cleric"  ,
                2: "assassin",
                3: "knight"  ,
                4: "ranger"  ,
                5: "paladin" ,
                6: "monk"    ,
                7: "ninja"   ,
                8: "druid"   }

_genders = {0: "Neutral",
            1: "Male"   ,
            2: "Female" }

_item_types = {0 : 'Undefined'    ,
               1 : 'Light'        ,
               2 : 'Scroll'       ,
               3 : 'Wand'         ,
               4 : 'Staff'        ,
               5 : 'Weapon'       ,
               6 : 'RangeWeapon'  ,
               7 : 'Missile'      ,
               8 : 'Treasure'     ,
               9 : 'Armor'        ,
               10: 'Potion'       ,
               11: 'Worn'         ,
               12: 'Other'        ,
               13: 'Trash'        ,
               14: 'Trap'         ,
               15: 'Container'    ,
               16: 'Note'         ,
               17: 'LiquidContainer',
               18: 'Key'          ,
               19: 'Food'         ,
               20: 'Money'        ,
               21: 'Pen'          ,
               22: 'Boat'         ,
               23: 'Fountain'     ,
               24: 'Portal'       ,
               25: 'Pill'         }

# Extra what? (item)
_extra_bits = {0 : 'Glowing'       ,
               1 : 'Humming'       ,
               2 : 'NoRent'        ,
               3 : 'NoDonate'      ,
               4 : 'NoInvis'       ,
               5 : 'Invisible'     ,
               6 : 'Magic'         ,
               7 : 'Cursed'        ,
               8 : 'Blessed'       ,
               9 : 'AntiGood'      ,
               10: 'AntiEvil'      ,
               11: 'AntiNeutral'   ,
               12: 'TwoHanded'     ,
               13: 'MagicallyCreated',
               14: 'Flaming'       ,
               15: 'Poison'        ,
               16: 'NoSell'        ,
               17: 'NoMobTake'     ,
               18: 'Quest'         ,
               19: 'NoDecay'       ,
               20: 'NoLocate'      ,
               21: 'Donated'       ,
               22: 'NoAuction'     ,
               23: 'Shining'       ,
               24: 'Unapproved'    ,
               25: 'Hidden'        ,
               26: 'NoDisarm'      ,
               27: 'Transparent'   ,
               28: 'Biodegradable' ,
               29: 'Deleted'       ,
               30: 'NoIdentify'    ,
               31: 'ItemDecaying'  }
               # There aren't 33 bits in a bitvector.
               # 32: 'ItemNoDecay'   }

_anticlass_bits = {0: 'wizard'  ,
                   1: 'cleric'  ,
                   2: 'assassin',
                   3: 'knight'  ,
                   4: 'ranger'  ,
                   5: 'paladin' ,
                   6: 'monk'    ,
                   7: 'ninja'   ,
                   8: 'druid'}

_apply_types = {0 : 'None'                   ,
                1 : 'Strength'               ,
                2 : 'Dexterity'              ,
                3 : 'Intelligence'           ,
                4 : 'Wisdom'                 ,
                5 : 'Constitution'           ,
                6 : 'Charisma'               ,
                7 : 'Class'                  ,
                8 : 'Level'                  ,
                9 : 'Age'                    ,
                10: 'Weight'                 ,
                11: 'Height'                 ,
                12: 'MaxMana'                ,
                13: 'MaxHit'                 ,
                14: 'MaxMove'                ,
                15: 'Gold'                   ,
                16: 'Experience'             ,
                17: 'Armour'                 ,
                18: 'Hitroll'                ,
                19: 'Damroll'                ,
                20: 'Saving vs. Paralysis'   ,
                21: 'Saving vs. Rod/Staff/Wand',
                22: 'Saving vs. Petrification',
                23: 'Saving vs. Breath Weapon',
                24: 'Saving vs. Spells'      }

_container_bits = {0: 'Closeable',
                   1: 'Pickproof',
                   2: 'Closed' ,
                   3: 'Locked' }

_container_types = {0: 'Normal',
                    1: 'Corpse',
                    2: 'Pile',
                    3: 'Vomit'}

_drinks = {0 : 'water'         ,
           1 : 'beer'          ,
           2 : 'wine'          ,
           3 : 'ale'           ,
           4 : 'dark ale'      ,
           5 : 'whisky'        ,
           6 : 'lemonade'      ,
           7 : 'firebreather'  ,
           8 : 'local speciality',
           9 : 'slime mold juice',
           10: 'milk'          ,
           11: 'tea'           ,
           12: 'coffee'        ,
           13: 'blood'         ,
           14: 'salt water'    ,
           15: 'clear water'   ,
           16: 'champagne'     ,
           17: 'eggnog'        ,
           18: 'fireoil'       ,
           19: 'embalming fluid'}

_drinknames = {0 : 'water'     ,
               1 : 'beer'      ,
               2 : 'wine'      ,
               3 : 'ale'       ,
               4 : 'ale'       ,
               5 : 'whisky'    ,
               6 : 'lemonade'  ,
               7 : 'firebreather',
               8 : 'local'     ,
               9 : 'juice'     ,
               10: 'milk'      ,
               11: 'tea'       ,
               12: 'coffee'    ,
               13: 'blood'     ,
               14: 'salt'      ,
               15: 'water'     ,
               16: 'champagne' ,
               17: 'nog'       ,
               18: 'oil'       ,
               19: 'fluid'     }

_color_liquid = {0 : 'clear'              ,
                 1 : 'brown'              ,
                 2 : 'clear'              ,
                 3 : 'brown'              ,
                 4 : 'dark'               ,
                 5 : 'golden'             ,
                 6 : 'red'                ,
                 7 : 'green'              ,
                 8 : 'clear'              ,
                 9 : 'light green'        ,
                 10: 'white'              ,
                 11: 'brown'              ,
                 12: 'black'              ,
                 13: 'red'                ,
                 14: 'clear'              ,
                 15: 'crystal clear'      ,
                 16: 'thick yellowish white'}

_spell_wear_off_msg = {0 : 'RESERVED DB.C'                                              ,
                       1 : 'You feel less protected.'                                   ,
                       2 : '!Teleport!'                                                 ,
                       3 : 'You feel less righteous.'                                   ,
                       4 : 'You feel a cloak of blindness disolve.'                     ,
                       5 : '!Burning Hands!'                                            ,
                       6 : '!Call Lightning'                                            ,
                       7 : 'You feel more self-confident.'                              ,
                       8 : 'You feel your strength return.'                             ,
                       9 : 'The magical vines constricting you suddenly disappear.'     ,
                       10: '!Color Spray!'                                              ,
                       11: '!Control Weather!'                                          ,
                       12: '!Create Food!'                                              ,
                       13: '!Create Water!'                                             ,
                       14: '!Cure Blind!'                                               ,
                       15: '!Cure Critic!'                                              ,
                       16: '!Cure Light!'                                               ,
                       17: 'You feel more optimistic.'                                  ,
                       18: 'You feel less aware.'                                       ,
                       19: 'Your eyes stop tingling.'                                   ,
                       20: 'The detect magic wears off.'                                ,
                       21: 'The detect poison wears off.'                               ,
                       22: '!Dispel Evil!'                                              ,
                       23: '!Earthquake!'                                               ,
                       24: '!Enchant Weapon!'                                           ,
                       25: '!Energy Drain!'                                             ,
                       26: '!Fireball!'                                                 ,
                       27: '!Harm!'                                                     ,
                       28: '!Heal!'                                                     ,
                       29: 'You feel yourself exposed.'                                 ,
                       30: '!Lightning Bolt!'                                           ,
                       31: '!Locate object!'                                            ,
                       32: '!Magic Missile!'                                            ,
                       33: 'You feel less sick.'                                        ,
                       34: 'You feel less protected.'                                   ,
                       35: '!Remove Curse!'                                             ,
                       36: 'The white aura around your body fades.'                     ,
                       37: '!Shocking Grasp!'                                           ,
                       38: 'You feel less tired.'                                       ,
                       39: 'You feel weaker.'                                           ,
                       40: '!Summon!'                                                   ,
                       41: '!Ventriloquate!'                                            ,
                       42: '!Word of Recall!'                                           ,
                       43: '!Remove Poison!'                                            ,
                       44: 'You feel less aware of your suroundings.'                   ,
                       45: '!Animate Corpse!'                                           ,
                       46: '!Dispel Good!'                                              ,
                       47: '!Group Armor!'                                              ,
                       48: '!Group Heal!'                                               ,
                       49: '!Group Recall!'                                             ,
                       50: 'Your night vision seems to fade.'                           ,
                       51: 'Your feet seem less boyant.'                                ,
                       52: '!Identify!'                                                 ,
                       53: 'You float back down to the ground.'                         ,
                       54: 'The shield of flame around your body disappears.'           ,
                       55: 'Your whupass seems to have worn off.'                       ,
                       56: 'Your image stops shifting and returns to normal.'           ,
                       57: 'You feel solid again.'                                      ,
                       58: '!Create Spring!'                                            ,
                       59: '!Continual Light!'                                          ,
                       60: '!Teleport Object!'                                          ,
                       61: 'Your skin softens and returns to normal.'                   ,
                       62: 'Your barkskin returns to normal.'                           ,
                       63: '!Ice Storm!'                                                ,
                       64: '!Enchant Armor!'                                            ,
                       65: '!Portal!'                                                   ,
                       66: 'The force shield around you shimmers and fades away.'       ,
                       67: '!Refresh!'                                                  ,
                       68: 'Your feel your less energetic.'                             ,
                       69: '!Find the Path!'                                            ,
                       70: 'You feel your metabolism begin to slow down.'               ,
                       71: 'You feel less protected from magic.'                        ,
                       72: '!Earthmaw!'                                                 ,
                       73: 'Your magical gills glow briefly and disappear.'             ,
                       74: '!Flame Strike!'                                             ,
                       75: '!Conjure Fire Elemental!'                                   ,
                       76: '!Conjure Earth Elemental!'                                  ,
                       77: '!Conjure Water Elemental!'                                  ,
                       78: '!Conjure Air Elemental!'                                    ,
                       79: '!Energy Blast!'                                             ,
                       80: '!Ice Bolt!'                                                 ,
                       81: '!Restoration!'                                              ,
                       82: 'The green aura around you fades.'                           ,
                       83: '!Channel!'                                                  ,
                       84: '!Regain Mana!'                                              ,
                       85: 'The green aura around you fades.'                           ,
                       86: '!Mana Steal!'                                               ,
                       87: 'The calm peacefulness around you slowly recedes.'           ,
                       88: '!Divine Hammer!'                                            ,
                       89: '!Power Lance!'                                              ,
                       90: '!Turn!'                                                     ,
                       91: 'Your befuddlement suddenly disappears.  You blush sheepishly.',
                       92: '!UNUSED!'                                                   }

_weekdays = {0: 'the Day of the Moon'    ,
             1: 'the Day of the Bull'    ,
             2: 'the Day of the Deception',
             3: 'the Day of Thunder'     ,
             4: 'the Day of Freedom'     ,
             5: 'the Day of the Great Gods',
             6: 'the Day of the Sun'     }

_month_name = {0 : 'Month of Winter'            ,
               1 : 'Month of the Winter Wolf'   ,
               2 : 'Month of the Frost Giant'   ,
               3 : 'Month of the Old Forces'    ,
               4 : 'Month of the Grand Struggle',
               5 : 'Month of the Spring'        ,
               6 : 'Month of Nature'            ,
               7 : 'Month of Futility'          ,
               8 : 'Month of the Dragon'        ,
               9 : 'Month of the Sun'           ,
               10: 'Month of the Heat'          ,
               11: 'Month of the Battle'        ,
               12: 'Month of the Dark Shades'   ,
               13: 'Month of the Shadows'       ,
               14: 'Month of the Long Shadows'  ,
               15: 'Month of the Ancient Darkness',
               16: 'Month of the Great Evil'    }

_sun_types = {0: '&BNight&N',
              1: '&YSunrise&N',
              2: '&yDaytime&N',
              3: '&YSunset&N'}

_sky_types = {0: '&wClear Skies&N',
              1: 'Cloudy'       ,
              2: '&bRaining&N'  ,
              3: '&yLightning&N'}

UNUSED    = '!UNUSED!'
UNDEFINED = '!UNDEFINED!'
RESERVED  = '!RESERVED!'

_tunnel_size = 2


# Conversion Utilities.
##    import re
##    PATTERN = re.compile(r'#define\s+(\w+)(?:\s+)(.*)')
##
##    def convert_defines(config):
##        config.replace('\t', '    ')
##        for line in config.split('\n'):
##            line = line.rstrip()
##            match = PATTERN.match(line)
##            if match is None:
##                print line
##            else:
##                (name, value) = match.groups()
##                pt = value.find('/*')
##                if pt >= 0:
##                    comment = value[pt+2:]
##                    value = value[:pt].rstrip()
##
##                    pt = comment.rfind('*/')
##                    if pt >= 0:
##                        comment = comment[:pt]
##
##                    comment = '# ' + comment.strip()
##                else:
##                    comment = ''
##
##                print '%-30s = %-20s%s' % (name, value, comment)
##
##    def convert_arrays(these):
##        from sys import stdout
##        for (name, array) in []:
##            longest = max(len(i) for i in array)
##            ln = len(array)
##
##            fmt = '%%-%dd: %%-%dr' % (len(str(ln)), longest)
##            header = '%s = {' % name
##            tab = ',\n' + len(header) * ' '
##
##            stdout.write(header)
##            for x in xrange(ln):
##                if x:
##                    stdout.write(tab)
##
##                stdout.write(fmt % (x, array[x]))
##
##            stdout.write('}\n\n')
##
##    if __name__ == '__main__':
##        convert_defines(CONFIG)



# The WRLC subsystem interoperates with native functionality, but because ph is
# higher level, many of these things exist for symbolic compatibility with the
# structural model ported from WRLC.  The structural module should be modified
# to use ph components.


from contextlib import contextmanager
# from types import DictionaryType, ClassType, DictProxyType, TypeType, InstanceType, ModuleType, CodeType
DictionaryType = dict
from types import new_class as ClassType
from types import ModuleType, CodeType

import re

def dictOf(o):
    return o.__dict__
def classOf(o):
    return o.__class__

def apply(function, *args, **kwd):
    return function(*args, **kwd)


# Currently unsupported:
# docstring

# Document.Section:
#     GenericSearch.Results.{points,selected}
#     search

# Core:
#     Objects:
#         encoded
#         url
#         format

#     Document:
#         extraction
#         processing


def getPathType():
    # The only real WRLC dependency.
    # The filesystem interface might be useful, but should
    # probably be put into the ph.components package if you
    # don't want this feature to fail.
    from common.platform.path import PathType
    return PathType


# Stubs.
class Object:
    class Format:
        def __init__(self, name):
            pass
        def __call__(self, object):
            return object

# class GenericContext:
#     def __enter__(self):
#         return self

#     def __exit__(self, etype = None, value = None, tb = None):
#         if etype is not None:
#             raise etype, value, tb


class Namespace(dict):
    @classmethod
    def ToStructure(self, object):
        # That is, to primitive type (not object.
        if isinstance(object, (dict, DictProxyType)):
            return dict((n, self.ToStructure(v)) \
                        for (n, v) in object.items())

        if isinstance(object, (ClassType, TypeType, InstanceType)):
            return self.ToStructure(dictOf(object))

        if isinstance(object, list):
            return list(map(self.ToStructure, object))
        if isinstance(object, tuple):
            return tuple(self.ToStructure(o) for o in object)

        return object

    @classmethod
    def FromStructure(self, structure, strictTypes = True):
        if strictTypes:
            def typeCompare(i, *types):
                return type(i) in types
        else:
            def typeCompare(i, *types):
                return isinstance(i, types)

        def _(v):
            # Namespaces are already built?  This means that you should wrap
            # anything that might recurse like so.
            if not isinstance(v, Namespace):
                if typeCompare(v, dict):
                    return self.FromStructure(v)

                if typeCompare(v, tuple):
                    return tuple(_(i) for i in v)

                if typeCompare(v, list):
                    return list(map(_, v))

                # Todo: what about generator sequences?

            return v

        # Ostensibly the same as the (implicit) recursor,
        # except make sure to look at the...
        if typeCompare(structure, tuple):
            return tuple(self.FromStructure(v) for v in structure)
        if typeCompare(structure, list):
            return list(map(self.FromStructure, structure))

        if typeCompare(structure, dict):
            # ...dictionary-type reaction:
            kwd = dict((name, _(value)) for (name, value) \
                       in structure.items())
            return self(kwd) # **kwd)

        return structure

    # Todo: mutating deep-set

    def __init__(self, sequence = None, **kwd):
        # Q: Err, derive from dict? really?
        if isinstance(sequence, dict):
            sequence = iter(sequence.items())

        if sequence:
            for pair in sequence:
                if isinstance(pair, (list, tuple)) and len(pair) == 2:
                    (name, value) = pair
                    self[name] = value

        self.update(kwd)
        self.__dict__ = self

class Synthetic(Object):
    # This should implement partials/mixins
    def __init__(self, dict = None, **values):
        if not isinstance(dict, DictionaryType):
            assert dict is None
            dict = values
        else:
            dict.update(values)

        self.__dict__ = dict

    def __getitem__(self, item):
        return self.__dict__.__getitem__(item)
    def __setitem__(self, item, value):
        self.__dict__.__setitem__(item, value)

    def getAttributeString(self):
        from reprlib import repr
        # XXX :skip: fail on recursively-referencing structures.
        return ', '.join('%s=%s' % (n, repr(v)) for (n, v) in \
                         self.__dict__.items())

    def __str__(self):
        return inspect(dictOf(self))

class GenericContext:
    def __enter__(self):
        return self
    def __exit__(self, etype = None, evalue = None, tb = None):
        if evalue is not None:
            raise etype(evalue).with_traceback(tb)


# class Namespace(dict):
#     def __init__(self, *args, **kwd):
#         dict.__init__(self, *args, **kwd)
#         self.__dict__ = self

# class Synthetic(Namespace):
#     pass


def ExecuteText(source, locals = None, globals = None,
                codeSource = '<interactive text input>'):

    if source is not None:
        if isinstance(source, str):
            code = compile(source, codeSource, 'exec')
        else:
            code = source
            assert isinstance(code, CodeType), ValueError(code)

        if locals is None:
            locals = dict()
        if globals is None:
            globals = locals

        exec(code, globals, locals)

    return (globals, locals)

exe = ExecuteText

class Executable:
    def executeIn(self, locals, globals, **kwd):
        ExecuteText(self.codeObject, locals = locals, globals = globals, **kwd)

    def newModule(self, name = None, values = None, **kwd):
        module = ModuleType(name or self.name)

        try: register = kwd.pop('registration')
        except KeyError:
            pass
        else:
            # Register newly-created instance BEFORE execution,
            # (putting it in system.modules, for instances), so
            # that the code can access itself.
            if callable(register):
                register(module)

        ns = module.__dict__
        if isinstance(values, dict):
            ns.update(values)

        self.executeIn(ns, ns, **kwd)
        return module

    def execute(self, **kwd):
        import __main__ as main; main = main.__dict__
        self.executeIn(self.space, main, **kwd)

    def executeWith(self, values, **kwd):
        self.space.update(values)
        return self.execute()

    # __call__ = execute

    @property # memorized?
    def module(self):
        return self.newModule()

    def evaluate(self, locals = None, globals = None):
        if locals is None:
            locals = dict()
        if globals is None:
            globals = dict() # self.space

        return eval(self.codeObject, globals, locals)

    ##    def evaluate(self, *args, **kwd):
    ##        return eval(self.Original(self), *args, **kwd)

    evaluated = property(evaluate)

    @classmethod
    def Evaluate(self, code, **env):
        return self(code, name = env.pop('name')) \
               .evaluate(locals = env.pop('locals'),
                         globals = env.pop('globals'))


def splitOne(s, c):
    i = s.find(c)
    if i >= 0:
        return (s[:i], s[i+1:])

    return (s, '')


class Indent:
    class Buffer:
        @classmethod
        def Get(self, buf = None):
            if buf is None:
                return StringIO()

            return buf

def camelize(*parts):
    return ''.join(parts[0] + (p.capitalize() for p in parts[1:]))

def merge(b, **kwd):
    return dict(b, **kwd)

    # b.update(kwd)
    # return kwd


class SourceCode:
    def GenerateFunction(self, *args, **kwd):
        raise NotImplementedError

    class Function:
        def __init__(self, *args, **kwd):
            raise NotImplementedError


class ContextStackScope(Object, list): # UserDict
    @contextmanager
    def __call__(self, **variables):
        self.append(dict())

        try: yield self.update(variables)
        finally:
            self.pop()

    def getAttributeString(self):
        return '%d levels' % len(self)
 
    # Accessors.
    @property
    def current(self):
        return list.__getitem__(self, -1)

    # data = current # for UserDict, making the following unnecessary:

    def update(self, dict = None, **kwd):
        if dict is not None:
            self.current.update(dict)

        self.current.update(kwd)
        return self

    def __getitem__(self, name):
        # return self.current[name]

        e = None
        for i in range(len(self)-1, -1, -1):
            try: return list.__getitem__(self, i)[name]
            except KeyError as e: pass

        if e is None:
            e = KeyError(name)

        raise e

    ##    def __getattr__(self, name):
    ##        # XXX :skip: Need to simulate base-access order upon attribute.
    ##
    ##        try: return list.__getitem__(self, -1)[name]
    ##        except KeyError:
    ##            return list.__getattribute__(self, name)

    @property
    def parent(self):
        return self.Parent(self, len(self) - 2)

    class Parent:
        @property
        def parent(self):
            assert self.__level > 0
            return classOf(self)(self.__scope, self.__level - 1)

        @property
        def scope(self):
            return list.__getitem__(self.__scope, self.__level)

        def __init__(self, scope, level):
            self.__scope = scope
            self.__level = level

        def __getitem__(self, name):
            return self.scope[name]
        def keys(self):
            return list(self.scope.keys())


# class ContextStackScope(list):
#     @contextmanager
#     def __call__(self, **kwd):
#         self.append(kwd)

#         try: yield
#         finally:
#             self.pop()


#     def __getattr__(self, name):
#         try: return object.__getattribute__(self, name)
#         except AttributeError:
#             for s in reversed(self):
#                 try: return s[name]
#                 except KeyError:
#                     continue

#             raise AttributeError(name)


@contextmanager
def mutateObject(object, **values):
    prev = dict()

    for (name, v) in values.items():
        try: prev[name] = getattr(object, name)
        except AttributeError: pass

        setattr(object, name, v)

    try: yield
    finally:
        for (name, v) in prev.items():
            setattr(object, name, v)


def registerSystemModule(name, module):
    from sys import modules
    modules[name] = module
    return module


def LookupObject(name):
    if ':' in name:
        (module, cls) = name.split(':')

        object = __import__(module, fromlist = [''])

        for n in cls.split('.'):
            object = getattr(object, n)

        return object

    return __import__(name, fromlist = [''])


WS = re.compile(r'^(\s*)').match
def indentOf(n):
    m = WS(n)
    return 0 if m is None else len(m.group(1))

def dedentBy(lines, i, lo = 0, hi = None):
    if hi is None:
        hi = len(lines)

    for x in range(lo, hi):
        n = lines[x]
        for c in n[:i]:
            if c not in ' \t':
                raise SyntaxError(repr(c))

        lines[x] = n[i:]

def dedent(string):
    lines = string.split('\n')

    # Delete any leading blank lines.
    for f in range(len(lines)):
        if lines[f]:
            break

    del lines[:f]

    if lines:
        i = indentOf(lines[0])
        dedentBy(lines, i)
        lines.append('')

    return '\n'.join(lines)



# MUD Player HTTP Module.
HTTP_REDIRECT_MESSAGE = \
'''HTTP/1.0 302 Found
Location: %(location)s
Connection: close

No page content to render.
'''

HTTP_REDIRECT_VARS = dict(location = 'http://localhost/stuph/')
CON_CLOSE = 'Disconnecting'

def getHttpRedirectResponse():
    # date = 'Thu, 11 Feb 2010 18:18:44 GMT'
    vars = HTTP_REDIRECT_VARS.copy()

    # Ugh, this whole darn thing could come from config..
    from ph import getConfig
    url = getConfig('http-redirect-url') or ''
    url = url.strip()
    if url:
        vars['location'] = url

    response = HTTP_REDIRECT_MESSAGE % vars
    response = response.replace('\n', '\r\n')

    return response

def handleHttpRedirect(peer, xxx_todo_changeme):
    (method, resource, protocol) = xxx_todo_changeme
    from ph import log as mudlog
    mudlog('HTTP LOGIN REDIRECT [%s]: %r' % (protocol, resource))

    # Send redirect response.  Disconnect.
    peer.write(getHttpRedirectResponse())
    peer.state = CON_CLOSE

    return True

def detectHttpRequest(peer, line):
    if peer.state != 'Get name':
        return False

    try: (method, resource, protocol) = line.split()
    except ValueError:
        return False

    if method.upper() not in ['GET', 'POST', 'OPTIONS', 'HEAD']:
        return False
    if not protocol.upper().startswith('HTTP'):
        return False

    return handleHttpRedirect(peer, (method, resource, protocol))



# MUD Runtime -- Event Objects.
# Todo: from the component library/support module.
from stuphos.etc import getLetters, listSortInsert, logException

from sys import exc_info as getExceptionInfo
from traceback import print_exception as printException

# Event Handler Model.
class EventResult(Exception):
    def __init__(self, result):
        self.result = result
        # raise self

def eventResult(result):
    raise EventResult(result)

class MutableInstanceMethodObject:
    # Hack to be able to snap other members onto it.
    def __init__(self, method):
        self.__method = method
        self.__repr__ = method.__repr__
        self.__name__ = method.__name__
        self.__doc_ = method.__doc__

    def __call__(self, *args, **kwd):
        return self.__method(*args, **kwd)

class EventController:
    # Prismatic.
    def __init__(self, event_name):
        self.event_name = event_name
        self.handlers = []

        # Because: See Binding.__getattr__
        self.decorator = MutableInstanceMethodObject(self.decorator)
        self.decorator.last = self.last
        self.decorator.first = self.first

        self.on = self.decorator

    def __call__(self, *args, **kwd):
        for h in self.handlers:
            try: self.callHandler(h, *args, **kwd)
            except EventResult as e:
                return e.result

    def callHandler(self, handler, *args, **kwd):
        try: return handler(self, *args, **kwd)
        except EventResult:
            raise
        except:
            self.handleException(*getExceptionInfo())

    def handleException(self, cls, value, tb):
        # printException
        logException(cls, value, tb, traceback = True)

    def registerHandler(self, handler):
        # Facilitate module reloading by removing all equivilant handlers.
        removal = []
        for h in self.handlers:
            if h == handler:
                removal.append(h)

        for h in removal:
            self.handlers.remove(h)

        # Now insert via priority.
        listSortInsert(self.handlers, handler,
                       key = lambda h: h.priority)

    def __iadd__(self, handler):
        self.registerHandler(handler)
        return self

    def decorator(self, function):
        self.registerHandler(Concretion(function))
        return function

    def last(self, function):
        self.registerHandler(Concretion(function, priority = HandlerBase.PRIORITY_LAST))
        return function

    def first(self, function):
        self.registerHandler(Concretion(function, priority = HandlerBase.PRIORITY_FIRST))
        return function

    def __repr__(self):
        return '%s(%s): %d handlers' % (self.__class__.__name__, self.event_name, len(self.handlers))

class Binding(object):
    def __init__(self, module):
        self.module = module

    def getController(self, name):
        try: ctlr = getattr(self.module, name)
        except AttributeError:
            ctlr = EventController(name)
            setattr(self.module, name, ctlr)
        else:
            assert issubclass(ctlr.__class__, EventController)

        return ctlr

    def __getattr__(self, name):
        try: return object.__getattribute__(self, name)
        except AttributeError as e:
            if not name.startswith('_'):
                # See EventController.__init__
                return self.getController(name).decorator

            raise e


    # def __getattr__(self, name):
    #     # XXX won't this recurse on self.getController? :skip:
    #     # todo: first try to do an object.__getattribute__, catching AttributeError and then do the controller decorator
    #     if not name.startswith('_'):
    #         # See EventController.__init__
    #         return self.getController(name).decorator

    #     # XXX object.__getattribute__ :skip:
    #     return object.__getattr__(self, name)

def declareEventController(name, **others):
    controllerClass = others.get('Controller', EventController)
    Module = others.get('Module')

    ctlr = controllerClass(name)
    if type(Module) is str:
        try: Module = __import__(Module)
        except ImportError:
            Module = None

    if Module is not None:
        setattr(Module, name, ctlr)

    return ctlr

def declare(module, events):
    for name in events:
        if not hasattr(module, name):
            ctlr = EventController(name)
            setattr(module, name, ctlr)

class EventBridge:
    # this is not used but a similar symbol is defined in system/__init__.
    class Calling:
        def __init__(self, name):
            self.name = name
        def __call__(self, *args, **kwd):
            try: from stuphos.system.api import game
            except ImportError:
                from stuphos.system import game

            bridgeModule = game.bridgeModule()
            if bridgeModule:
                function = getattr(bridgeModule, self.name, None)
                if callable(function):
                    return function(*args, **kwd)

        def __repr__(self):
            print('Calling: %r' % self.name)
        __str__ = __repr__

    def __setattr__(self, name, value):
        self.__dict__[name] = self.Calling(value)

# Handler Types.
class HandlerBase:
    PRIORITY_FIRST = 0
    PRIORITY_LAST = 10000
    PRIORITY_NORMAL = (PRIORITY_LAST - PRIORITY_FIRST) / 2

    def __init__(self, priority = None):
        self.priority = priority

    def __call__(self, ctlr, *args, **kwd):
        pass

    def sameClass(self, other):
        try:
            thisClass = self.__class__
            otherClass = other.__class__

            return thisClass.__module__ == otherClass.__module__ and \
                   thisClass.__name__   == otherClass.__name__

        except AttributeError:
            return False

    def __eq__(self, other):
        return self.sameClass(other)
    def __cmp__(self, other):
        return cmp(self.priority, other.priority)

    @property
    def priority(self):
        return getattr(self, '_priority', self.PRIORITY_NORMAL)

    @priority.setter
    def priority(self, value):
        self._priority = self.PRIORITY_NORMAL \
                         if value is None \
                            else int(value)

class MemberResolutionHandler(HandlerBase):
    def __init__(self, method_name):
        method_name = method_name.split()[0].split('.')

        self.module = '.'.join(method_name[:-1])
        self.method = method_name[-1]

    def __call__(self, ctlr, *args, **kwd):
        try: module = self._import_module(self.module)
        except ImportError: pass
        else:
            if module:
                method = getattr(module, self.method, None)
                if callable(method):
                    return method(*args, **kwd)

    def _import_module(self, name):
        # Todo: import parts of name, in case the trailing items
        # are not submodules, but members.
        # (see mud/__init__: world.heartbeat.pulse)
        module = __import__(name)
        for m in name.split('.')[1:]:
            module = getattr(module, m)

        return module

    def __eq__(self, other):
        return self.sameClass(other) and \
               self.module == other.module and \
               self.method == other.method

    def __repr__(self):
        return 'resolve %s.%s' % (self.module, self.method)

class LoggingHandler(HandlerBase):
    def logEvent(self, ctlr, *args, **kwd):
        tab = '\n\t\t\t\t'

        msg = tab.join(map(repr, args))
        msg = '%s:%s%s' % (ctlr.event_name, tab, msg)

        if kwd:
            from pprint import pformat
            msg += tab
            msg += pformat(kwd)

        from stuphos import log
        log(msg)

    def __call__(self, ctlr, *args, **kwd):
        self.logEvent(ctlr, *args, **kwd)

class Concretion(HandlerBase):
    def __init__(self, function, *args, **kwd):
        HandlerBase.__init__(self, *args, **kwd)
        self.function = function

    def __call__(self, ctlr, *args, **kwd):
        # XXX todo: curry ctlr?? :skip:
        return self.function(*args, **kwd)

    def __eq__(self, other):
        if self.sameClass(other):
            return self.function.__module__ == other.function.__module__ and \
                   self.function.__name__ == other.function.__name__

    def __repr__(self):
        return repr(self.function)

# Parsing.
COMMENT_CHAR  = '#'
SECTION_LEFT  = '['
SECTION_RIGHT = ']'

def parseEvents(events, handler = None):
    for line in events.split('\n'):
        line = line.strip()
        if not line or line[0] == COMMENT_CHAR:
            continue

        if line[0] == SECTION_LEFT and line[-1] == SECTION_RIGHT:
            handler = line[1:-1]
        else:
            ws = line.find(' ')
            if ws < 0:
                yield line, handler, ''
            else:
                yield line[:ws], handler, line[ws + 1:].lstrip()

# A Configuration Layer.
_builtin_handlers = {'MemberResolution': MemberResolutionHandler,
                     'Logging'         : LoggingHandler}

def loadOnto(events, module):
    # Algorithm.
    for (name, handler, method) in parseEvents(events):
        ctlr = getattr(module, name, None)
        if ctlr is None:
            ctlr = EventController(name)
            setattr(module, name, ctlr)

        handlerClass = getHandlerClass(handler)
        if issubclass(handlerClass, HandlerBase):
            if method:
                ctlr += handlerClass(method)
            else:
                ctlr += handlerClass()

def getHandlerClass(name):
    return _builtin_handlers.get(getLetters(name))

def getEventNames(module):
    try: return module.implements
    except AttributeError: pass

    # Broad-spectrum:
    from . import EVENT_NAMES
    # import pdb; pdb.set_trace()
    return EVENT_NAMES



# Copyright 2011 Clint Banis.  All rights reserved.
#

from io import StringIO as new_buffer

# Formats.
# import cPickle as pickle
from pickle import dump as save_pickle, dumps as save_pickle_string
from pickle import load as load_pickle, loads as load_pickle_string

def serialize_integer(n):
    return 'I%d.' % int(n)

class UnknownLengthMarker(Exception):
    pass

def deserialize_integer(buf):
    c = buf.read(1)
    if c != 'I':
        raise UnknownLengthMarker(c)

    digits = []
    while True:
        c = buf.read(1)
        if c == '.':
            break

        assert c.isdigit()
        digits.append(c)

    assert digits
    return int(''.join(digits))

# Message Marshalling.
MAX_RAW_INPUT_LENGTH = 512
MAX_SUBUFFER_SIZE    = MAX_RAW_INPUT_LENGTH

PACKET_SIZE          = MAX_SUBUFFER_SIZE - 10

# XXX Escape telnet sequences!
# if IAC in data:
#    data = data.replace(IAC, IAC + IAC)
def prepare_sudata_segments(data, packet_size = PACKET_SIZE):
    # Telnet processing segments overflow data automatically, but it's convenient
    # to packetize it here in the client to control bandwidth.
    assert type(data) is str
    datalength = len(data)

    n = serialize_integer(datalength)
    s = 0

    if packet_size <= 0:
        yield '%s%s' % (n, data)
        return # StopIteration?

    # First packet will be offset by the message size overhead.
    e = packet_size - len(n)
    yield '%s%s' % (n, data[:e])

    s = e
    e += packet_size

    while s < datalength:
        yield data[s:e]
        s = e
        e += packet_size
        if e > datalength:
            e = datalength

from telnetlib import IAC, SB, SE
def embed_segment(part):
    return '%s%s%s%s%s' % (IAC, SB, part, IAC, SE)

def send_sudata(sock, data, **kwd):
    for part in prepare_sudata_segments(data, **kwd):
        sock.sendall(embed_segment(part))

def send_supickle(sock, object, **kwd):
    send_sudata(sock, save_pickle_string(object), **kwd)

# Message Unmarshalling.
class SuMessageBuilder:
    def __init__(self):
        self.readq = new_buffer()
        self.message_buf = None

        self.message_size = 0
        self.message_left = 0

    def receive_command(self, cmd):
        if type(cmd) is tuple:
            # Filter for SE messages.
            if len(cmd) == 3 and cmd[0] == ord(SE):
                return self.receive_sudata(cmd[2])

    # Subnegotiation message building.
    def push_data(self, data):
        pos = self.readq.tell()
        self.readq.write(data)
        self.readq.seek(pos)

    def read_integer(self):
        return deserialize_integer(self.readq)
    def read_message(self, bytes):
        return self.readq.read(bytes)

    def receive_sudata(self, data):
        self.push_data(data)

        if self.message_buf is None:
            self.message_size = self.read_integer()
            assert self.message_size >= 0
            self.message_buf = new_buffer()
            self.message_left = self.message_size

        data = self.read_message(self.message_left)

        self.message_left -= len(data)
        self.message_buf.write(data)

        # Todo: If nothing's left at the end of the readq, truncate it.

        if not self.message_left:
            return self.reset_message()

    def reset_message(self):
        buf = self.message_buf
        self.message_buf = None
        buf.reset()
        # Reset message_size/left?
        return buf

TelnetConversation = SuMessageBuilder



# Adapter entity types.
from stuphos.triggers import EntityAdapter, TriggerAdapter
from stuphos.kernel import BypassReturn, vmCurrentTask, Continuation, accessItemsNative, nullproc
from stuphos.runtime.architecture import writeprotected
# from stuphos.etc.tools.misc import accessItems as _accessItems

from sys import exc_info

class EntityNoAccessException(AttributeError):
    pass

class InlineObject:
    def _getWrappedAttribute(self, name):
        from world import heartbeat as vm
        if vm.isPulseExecuting:
            return getattr(self._object, name)

        return self._getAttributeForTask(vm, vmCurrentTask(), name)

    def _getAttributeForTask(self, vm, task, name):
        # Asynchronous message passing of getattr to heartbeat.
        st = vm.suspendedTasks

        def getAttribute():
            # Todo: use vm.threadContext and outer frame completion for dynamic getAttr$?
            # No: if this is an inline object, a wrapped attribute should always be layer
            # zero.
            try: value = getattr(self._object, name)
            except:
                st.resumeTask(task, vm, exception = exc_info())
            else:
                st.resumeTask(task, vm, value = value)

        st.suspendTask(task)
        vm.enqueueHeartbeatTask(getAttribute)

        raise BypassReturn

class perspective(EntityAdapter):
    # Contains perspective.
    # Todo: rename to yet another class named 'scoped'
    def __init__(self, object, perspective = None):
        self._perspective = perspective
        EntityAdapter.__init__(self, object)

    def _rootPerspective(self):
        if self._perspective is None:
            return self

        return self._perspective._rootPerspective()

    def _perspectiveIs(self, types):
        return isinstance(self._perspective, types)
    _isPerspective = _perspectiveIs

    def _perspectiveCheck(self, types, *args, **kwd):
        if not self._isPerspective(types):
            exceptionClass = kwd.pop('exceptionClass', AttributeError)
            raise exceptionClass(*args, **kwd)


class verbal:
    # Implements verbs.

    # @property
    # def verbs(self):
    #     return self._object.verbs
    # @property
    # def interface(self):
    #     return self._object.interface


    def verbCallObject(self, name):
        'Do command namespace lookup and return verb action.'

        # debugOn()
        try:
            i = self._object._interface
            a = self._object._verbs.abbrevs[name] # areas/generic/roomVerbs

            if isinstance(a, str):
                a = a.split('.')

            # Do effective swapFrameCall
            task = vmCurrentTask()
            # task -= task.frame

            # return accessItemsNative(task.frame, i.commands, *a)

            i._invoke(task.frame, dict(), 'lookup$', *['commands'] + a)

        except (AttributeError, KeyError):
            pass


class PeerExtensionPoint(perspective):
    _readonly_properties = ['state', 'prompt', 'host']
    __public_members__ = ['input', 'output']

    # MSTR(   last_host),
    # MSTR(  olc_string),
    # // MSTR(       inbuf),
    # MSTR(  last_input),
    # MSTR(small_outbuf),

    # // Fraun Sep 30th 2005 - Fixed these.  Was not T_STRING_INPLACE
    # M(email,  T_STRING, "email"),
    # M(remort, T_STRING, "remort"),

    # // See clear_output.
    # M(output, T_STRING, "output"),

    # MINT(  descriptor),
    # MINT(  ident_sock),
    # MINT(   peer_port),
    # MINT(          ip),
    # MINT(       login),
    # MINT(   connected),
    # MINT(    desc_num),
    # MINT(  login_time),
    # // MINT(     mail_to),
    # MINT(   mail_mode),
    # MINT(  has_prompt),
    # MINT(  color_mode),
    # MINT(  medit_mode),
    # MINT(  oedit_mode),
    # MINT(  sedit_mode),
    # MINT(  zedit_mode),
    # MINT(  zedit_page),
    # MINT( zedit_index),
    # MINT(  hedit_mode),
    # MINT(  cedit_mode),
    # MINT(   vt_bar_on),
    # MINT(       saved),
    # MINT(affect_index),
    # MINT(   edit_vnum),
    # MINT(   edit_room),
    # MINT(   edit_zone),
    # MINT(   edit_clan),
    # MINT(  help_entry),
    # MINT(   redit_dir),
    # MINT(  redit_mode),
    # MINT(     olc_int),
    # MINT( history_pos),
    # MINT(      bufptr),
    # MINT(    bufspace),

    # MINT(last_command_time),
    # MINT(    olc_save_type),
    # MINT(     bandwidth_in),
    # MINT(    bandwidth_out),

    # // Fraun Sep 30th 2005 - Not T_STRING_INPLACE
    # M(showstr_head,  T_STRING, "showstr_head"),
    # M(showstr_point, T_STRING, "showstr_point"),
    # M(showstr_start, T_STRING, "showstr_start"),

    # // $!@@# showstr_vector is (char **) !@#$@!
    # // MSTR(showstr_vector),

    # MINT( showstr_count),
    # // MINT(showstr_vector),
    # MINT(       max_str),

    def sendln(self, message):
        'peer.sendln("text string message without newline")'

        # debugOn()
        return self._object.sendln(message)


    @property
    def avatar(self):
        return MobileAdapter(self._object.avatar, perspective = self)

    @property
    def session(self):
        s = getattr(self._object, 'session_ref', None)
        if s is not None:
            s = s()
            if s is not None:
                return s.adapter

                # Return a new adapter instance?  No: cleanup on session is used during disconnection.
                from phsite.network.adapter.sessions import SessionAdapter
                return SessionAdapter(s)


    # interpreter = property(_get_interpreter, _set_interpreter)
    # messenger = property(_get_messenger, _set_messenger)

    @property
    def input(self):
        return self._object.input
    @input.setter
    def input(self, input):
        self._object.input = input

    @property
    def output(self):
        return self._object.output
    @output.setter
    def output(self, output):
        self._object.output = output

    def clearOutput(self):
        return self._object.clear_output()
    def editString(self, initial):
        return self._object.editString(initial)
    def next_command(self):
        return self._object.next_command()
    def page_string(self, string):
        return self._object.page_string(string)
    page = page_string

    def textout(self, text):
        return self._object.textout(text)
    def write(self, *args, **kwd):
        return self._object.write(*args, **kwd)
    def forceInput(self, input):
        return self._object.forceInput(input)
    def handleCommand(self, comm):
        return self._object.handleCommand(comm)


    # {"redirect", (PyCFunction)peer_redirect, METH_VARARGS, "A more broad/low-level form of intercepting output."},

    # {"members", (PyCFunction)peer_members, METH_NOARGS, ""},

    # {"reload",      (PyCFunction)peerReload,     METH_VARARGS,
    #  peerReloadDoc},

PeerAdapter = PeerExtensionPoint


class MobileExtensionPoint(InlineObject, perspective):
    @property
    def peer(self):
        peer = self._object.peer
        if peer is not None:
            # Permit object retrieval, but control its scope.
            return PeerAdapter(peer, perspective = self)

    @property
    def name(self):
        return self._object.name


    @property
    def location(self):
        room = self._object.location
        if room is not None:
            return RoomAdapter(room, perspective = self)

    @location.setter
    def location(self, value):
        if self._perspective is not None:
            raise EntityNoAccessException('location')
        if not isinstance(value, RoomAdapter):
            raise TypeError(type(value).__name__)

        self._object.location = value._object

    room = location


    # {"room", (getter)mobile_get_room, (setter)mobile_set_room},
    # {"location", (getter)mobile_get_room, (setter)mobile_set_room},
    # {"next", (getter)mobile_get_next},

    # {"zone", (getter)mobile_get_zone},

    # {"vnum",   (getter)mobile_get_vnum},

    # {"keywords", (getter)mobile_get_keywords, (setter)mobile_set_keywords},
    # {"short_descr", (getter)mobile_get_shortdescr, (setter)mobile_set_shortdescr},
    # {"long_descr", (getter)mobile_get_longdescr, (setter)mobile_set_longdescr},
    # {"description", (getter)mobile_get_description, (setter)mobile_set_description},

    # PYUSE_MOBGETSETPROP(sex),
    # PYUSE_MOBGETSETPROP(position),
    # PYUSE_MOBGETPROP(playerflags),
    # PYUSE_MOBGETPROP(npcflags),

    # // XXX Join-these:
    # PYUSE_MOBGETPROP(preferences),
    # PYUSE_MOBGETPROP(preferences2),

    # {"affectflags", (getter)mobile_get_affectvector},
    # PYUSE_MOBGETPROP(affectvector),
    # PYUSE_MOBGETPROP(affections),

    # // What about display-flags?  Maybe it should use mobile_display_flags defined in this file..

    # PYUSE_MOBGETPROP(equipment),
    # PYUSE_MOBGETPROP(inventory),

    # {"fighting", (getter)mobile_get_fighting, (setter)mobile_set_fighting},
    # {"opponent", (getter)mobile_get_fighting, (setter)mobile_set_fighting},

    # PYUSE_MOBGETSETPROP(master),
    # PYUSE_MOBGETPROP(holding),
    # PYUSE_MOBGETPROP(heldby),
    # PYUSE_MOBGETPROP(mount),
    # PYUSE_MOBGETPROP(mountedby),

    # PYUSE_MOBGETPROP(followers),
    # PYUSE_MOBGETPROP(colorset),

    # // {"gender", (getter)mobile_get_gender},
    # // {"sex",    (getter)mobile_get_gender},

    # // {"position", (getter)mobile_get_position},

    # // NPC Mobiles
    # {"__special__", (getter)mobile_get_specproc, (setter)mobile_set_specproc},
    # {"special",     (getter)mobile_get_specproc, (setter)mobile_set_specproc},
    # {"prototype",   (getter)mobile_get_prototype /* Settable! */},

    # {"specialName", (getter)mobile_get_specproc_name},

    # // Player
    # {"mortal",       (getter)mobile_get_ismortal},
    # {"immortal",     (getter)mobile_get_isimmortal},
    # {"god",          (getter)mobile_get_isgod},
    # {"supreme",      (getter)mobile_get_isimpl},
    # {"implementor",  (getter)mobile_get_isimpl},
    # {"isPlayer",     (getter)mobile_isPlayer},

    # // Or Non-player
    # {"npc",          (getter)mobile_get_isnpc},

    # {"alias",         (getter)mobile_get_allaliases,  (setter)mobile_set_allaliases},
    # // {"complex_alias", (getter)mobile_get_cmplxalias,  (setter)mobile_set_cmplxalias},
    # // {"simple_alias",  (getter)mobile_get_smplalias,   (setter)mobile_set_smplalias},

    # // {"check_mail",   (getter)mobile_get_mail},
    # // {"mail",         (getter)mobile_get_mail},
    # {"hasMail",      (getter)mobile_has_mail},

    # {"middleName",   (getter)mobile_get_middlename},

    # {"valid",        (getter)mobile_isvalid},
    # {"destructor",   (getter)mobile_get_destructor, (setter)mobile_set_destructor},

    # // Let's not define this right off: just set it on the dictionary straight.
    # // {"destroy",      (getter)mobile_get_destructor, (setter)mobile_set_destructor},

    # #ifdef EXPOSE_DICT
    #     {"__dict__",           (getter)mobile_get_dict},
    # #elif defined(EXPOSE_DICT_RENAMED)
    #     {EXPOSE_DICT_RENAMED,  (getter)mobile_get_dict},
    # #endif

    # {"power_rating", (getter)mobile_power_rating},
    # {"power",        (getter)mobile_power_rating},

    # {"programming",  (getter)mobile_get_programming, (setter)mobile_set_programming},
    # {"commands",     (getter)mobile_get_trusted_commands},
    # {"trusted",      (getter)mobile_get_trusted_commands},

    # {"armorclass",   (getter)mobile_get_armorclass},
    # {"armor",        (getter)mobile_get_armorclass},

    # {"wizinvis_level", (getter)mobile_get_wizinvis_level},

    # {"store_value",  (getter)mobile_get_playerstore},
    # {"playerstore",  (getter)mobile_get_playerstore},
    # {"store",        (getter)mobile_get_playerstore},

    # {"equipment",    (getter)mobile_get_equipment},

    # {"syslogLevel",  (getter)mobileGetSyslogLevel},

    # {"py_handle",    (getter)NULL, (setter)mobileSetHandle},
    # {"handle",       (getter)NULL, (setter)mobileSetHandle,

    # {"cmdlock",      (getter)mobileGetCmdlock, (setter)mobileSetCmdlock},

    # MINT(pfilepos   ), // int
    # MINT(nr         ), // mob_rnum  (unsigned int)
    # MINT(in_room    ), // room_rnum (unsigned int)
    # MINT(was_in_room), // room_rnum (unsigned int)

    # // Fraun Dec 3rd 2005 - Renamed from 'wait'
    # M(nr, T_INT, "waitState"),

    # { "passwd", T_STRING_INPLACE, offsetof(struct char_data, player.passwd), READONLY },

    # // Getset Descriptor
    # // MNSTR(player.name, "name"),

    # // MNSTR(player.short_descr, "short_descr"),
    # // MNSTR(player.long_descr, "long_descr"),
    # // MNSTR(player.description, "description"),

    # MNSTR(player.title, "title"),
    # MNSTR(player.nickname, "nickname"),
    # MNSTR(player.prename, "prename"),
    # MNSTR(player.wizname, "wizname"),
    # MNSTR(player.plan, "plan"),
    # MNSTR(player.email, "email"),

    # MNINT(player.remort, "remort"), // int -- descriptor?
    # MNINT(player.race, "race"),     // int -- descriptor?
    # MNINT(player.clan, "clan"),     // int -- descriptor?
    # MNINT(player.clanrank, "clanrank"),   // int
    # MNINT(player.save_room, "save_room"), // int
    # MNINT(player.page_length, "page_length"),  // int
    # MNINT(player.breath, "breath"), // int

    # // sex implemented as a descriptor
    # // chclass implemented as a descriptor
    # M(player.level, T_BYTE, "level"), // byte

    # MNINT(player.hometown, "hometown"), // int

    # // struct time_data time;

    # M(player.weight, T_UBYTE, "weight"), // ubyte
    # M(player.height, T_UBYTE, "height"), // ubyte

    # MNINT(player.login_time, "login_time"), // time_t

    # // Reimplement these as descriptors?
    # M(real_abils.str,     T_BYTE, "strength"),           // sbyte
    # M(real_abils.str_add, T_BYTE, "strength_addition"),  // sbyte
    # M(real_abils.intel,   T_BYTE, "intelligence"),       // sbyte
    # M(real_abils.wis,     T_BYTE, "wisdom"),             // sbyte
    # M(real_abils.dex,     T_BYTE, "dexterity"),          // sbyte
    # M(real_abils.con,     T_BYTE, "constitution"),       // sbyte
    # M(real_abils.cha,     T_BYTE, "charisma"),           // sbyte

    # M(aff_abils.str,      T_BYTE, "aff_strength"),           // sbyte
    # M(aff_abils.str_add,  T_BYTE, "aff_strength_addition"),  // sbyte
    # M(aff_abils.intel,    T_BYTE, "aff_intelligence"),       // sbyte
    # M(aff_abils.wis,      T_BYTE, "aff_wisdom"),             // sbyte
    # M(aff_abils.dex,      T_BYTE, "aff_dexterity"),          // sbyte
    # M(aff_abils.con,      T_BYTE, "aff_constitution"),       // sbyte
    # M(aff_abils.cha,      T_BYTE, "aff_charisma"),           // sbyte

    # M(points.mana, T_SHORT, "mana"),
    # M(points.max_mana, T_SHORT, "max_mana"),

    # M(points.hit, T_SHORT, "hit"),
    # M(points.max_hit, T_SHORT, "max_hit"),

    # // I'm dead serious.  See moveMobile, mobile_move, and NotifyMovement.  How about 'stamina?'
    # M(points.move, T_SHORT, "fatigue"),
    # M(points.max_move, T_SHORT, "max_fatigue"),

    # M(points.deaths, T_SHORT, "deaths"),
    # M(points.mkills, T_SHORT, "npckills"),
    # M(points.pkills, T_SHORT, "plrkills"), // inconsistancy with 'playerflags'?
    # M(points.dts, T_SHORT, "dts"),
    # M(points.qpoints, T_SHORT, "qpoints"),

    # // armorclass is computed.  See getset
    # M(points.armor, T_SHORT, "ac"),

    # MNINT(points.gold, "gold_on_hand"),
    # // MNINT(points.bank_gold, "gold_in_bank"),
    # MNINT(points.exp, "experience"),

    # M(points.hitroll, T_BYTE, "hitroll"),
    # M(points.damroll, T_BYTE, "damroll"),

    # MNSTR(char_specials.afk_message, "afk_message"),

    # // XXX Implement wielded_by -- or is this obselete? (temporary magic-casting mobiles as special weapons)

    # // position is implemented as a descriptor

    # MNINT(char_specials.carry_weight, "carry_weight"),
    # M(char_specials.carry_items, T_BYTE, "carry_items"), // byte

    # MNINT(char_specials.timer, "timer"),
    # MNINT(char_specials.arena, "arena"),
    # MNINT(char_specials.tracking, "tracking"),
    # MNINT(char_specials.track_timer, "track_timer"),

    # MNINT(char_specials.saved.alignment, "alignment"),
    # MNINT(char_specials.saved.idnum, "idnum"),

    # // Fraun Sep 30th 2005 - Access these two things through the descriptors
    # // MNINT(char_specials.saved.act, "flags"), // XXX obselete this
    # // MNINT(char_specials.saved.affected_by, "affected_bits"), // XXX obselete this

    # // XXX Implement player-specials separately:
    # //   - act and affected_by covered by npcflags and playerflags
    # //   - sh_int apply_saving_throw[5] getset descriptor

    # M(mob_specials.last_direction, T_BYTE, "last_dir"),

    # MNINT(mob_specials.attack_type, "attack_type"),
    # MNINT(mob_specials.walk_type, "walk_type"),

    # M(mob_specials.default_pos, T_BYTE, "default_pos"),
    # M(mob_specials.damnodice,   T_BYTE, "damnodice"),   // These could change to ints!
    # M(mob_specials.damsizedice, T_BYTE, "damsizedice"), // These could change to ints!

    # MNINT(mob_specials.timer, "timer"),

    # {"findplayer", (PyCFunction)mobile_findplayer, METH_VARARGS,
    #  mobile_findplayer_doc},
    # {"findchar",   (PyCFunction)mobile_findchar,   METH_VARARGS,
    #  mobile_findchar_doc},
    # {"finditem",   (PyCFunction)mobile_finditem,   METH_VARARGS,
    #  mobile_finditem_doc},

    # {"find",       (PyCFunction)mobile_generic_find, METH_VARARGS|METH_KEYWORDS,
    #  mobileGenericFindDoc},

    # // {"getalias",   (PyCFunction)mobile_getalias,   METH_VARARGS,
    # // mobile_getalias_doc},
    # // {"setalias",   (PyCFunction)mobile_setalias,   METH_VARARGS,
    # // mobile_setalias_doc},

    # {"members", (PyCFunction)mobile_members, METH_NOARGS, ""},

    # // ToDo: implement __contains__ to look in inventory and equipment list, recursing through containers.

    # {"wearing", (PyCFunction)mobileWearing, METH_VARARGS,
    #  "Lookup mobile equipment by position.  Do some other things too."},

    # {"instantiate", (PyCFunction)mobile_instantiate, METH_VARARGS|METH_KEYWORDS,
    #  mobile_instantiate_doc},
    # {"load",        (PyCFunction)mobile_instantiate, METH_VARARGS|METH_KEYWORDS,
    #  mobile_instantiate_doc},
    # {"new",         (PyCFunction)mobile_instantiate, METH_VARARGS|METH_KEYWORDS,
    #  mobile_instantiate_doc},
    # {"read_mobile", (PyCFunction)mobile_instantiate, METH_VARARGS|METH_KEYWORDS,
    #  mobile_instantiate_doc},

    # {"save",        (PyCFunction)mobile_save,        METH_NOARGS,
    #  "Calls save_player on PC/<write_mobproto_to_db> on NPC."},
    # {"unrent",      (PyCFunction)mobile_unrent,       METH_NOARGS,
    #  "Calls StuphAPI_Call(Crash_load) on PC."},
    # {"rent",        (PyCFunction)mobile_rent,         METH_VARARGS,
    #  "Calls Crash_rentsave."},

    # {"save_player_file", (PyCFunction)mobileSavePlayerFile, METH_VARARGS,
    #  ""},

    # {"move",        (PyCFunction)mobile_move,        METH_VARARGS,
    #  "Move a mobile from one room to another.  Same as setting 'room' or 'location'."},

    # {"update_position", (PyCFunction)mobile_update_position, METH_NOARGS, ""},
    # {"sever_limb", (PyCFunction)mobile_sever_limb, METH_VARARGS, ""},

    # {"make_corpse", (PyCFunction)mobile_make_corpse, METH_NOARGS, ""},
    # {"killAlignmentChange", (PyCFunction)mobile_kill_alignment_change, METH_VARARGS, ""},
    # {"deathCry", (PyCFunction)mobile_death_cry, METH_NOARGS, ""},
    # {"rawKill", (PyCFunction)mobile_raw_kill, METH_NOARGS, ""},
    # {"die", (PyCFunction)mobile_die, METH_NOARGS, ""},
    # {"stop_hunting", (PyCFunction)mobile_stop_hunting, METH_VARARGS, ""},
    # {"check_subdue", (PyCFunction)mobile_check_subdue, METH_VARARGS, ""},
    # {"incur_damage", (PyCFunction)mobile_incur_damage, METH_VARARGS, ""},
    # {"get_thac0", (PyCFunction)mobileGetTHAC0, METH_VARARGS,
    #  "Calculates the hit base for this mobile attacking given mobile parameter."},
    # {"attack", (PyCFunction)mobileAttack, METH_VARARGS, ""},
    # {"hit",    (PyCFunction)mobileAttack, METH_VARARGS, ""},
    # {"perform_attack", (PyCFunction)mobilePerformViolence, METH_VARARGS, ""},
    # {"attackModeOrder", (PyCFunction)mobileGetAttackModeOrder, METH_NOARGS, ""},

    # {"checkAutoassist", (PyCFunction)mobileCheckAutoAssist, METH_NOARGS, ""},
    # {"sameAs", (PyCFunction)mobileSameAs, METH_VARARGS, ""},
    # {"eq_apply_ac", (PyCFunction)mobile_eq_apply_ac, METH_VARARGS, ""},
    # {"invalid_item_align", (PyCFunction)mobile_invalid_item_align, METH_VARARGS, ""},

    # {"equip", (PyCFunction)mobileEquip, METH_VARARGS, ""},
    # {"wearing", (PyCFunction)mobileWearing, METH_NOARGS, ""},
    # {"unequip", (PyCFunction)mobileUnequip, METH_VARARGS, ""},

    # {"visibleToLevel", (PyCFunction)mobileVisibleToLevel, METH_VARARGS,
    #  mobileVisibleToLevelDoc},

    # {"visibleBy", (PyCFunction)mobileVisibleBy, METH_VARARGS,
    #  "Can the given mobile see this mobile?\n"},
    # {"canSee",    (PyCFunction)mobileCanSee,    METH_VARARGS,
    #  mobileCanSeeDoc},

    # {"do_auto_exits",  (PyCFunction)mobileDoAutoExits, METH_NOARGS,
    #  "Invokes do_auto_exits."},

    # {"changeHandle",   (PyCFunction)mobileChangeHandle, METH_VARARGS,
    #  "Returns the old reference to the old handle."},

    # {"act", (PyCFunction)mobileAct, METH_VARARGS | METH_KEYWORDS,
    #  mobileActDoc},

    # {"extract", (PyCFunction)mobileExtract, METH_NOARGS},

    # {"tell", (PyCFunction)mobilePerformTell, METH_VARARGS,
    #  "Perform a tell communication directly to target person."},
    # {"reply", (PyCFunction)mobilePerformReply, METH_VARARGS,
    #  "Reply to last target person."},

    # {"sayTo", (PyCFunction)mobilePerformSayTo, METH_VARARGS,
    #  "Perform a room-say directly to target (optional verb parameter at end)."},

    # {"consider", (PyCFunction)mobileConsider, METH_VARARGS,
    #  "Return consideration text between combatants.\n"},


    def sendln(self, message):
        return self._object.peer.sendln(message)

    def perform(self, command):
        return self._object.perform(command)
    __call__ = perform


    def interpretCommandInternal(self, command, arguments):
        return self._object.interpretCommandInternal(command, arguments)

    @property
    def securityContext(self):
        from phsite.network.adapter.sessions import SessionAdapter
        self._perspectiveCheck(SessionAdapter, 'Security context not available!')

        from ph.interpreter.mental.native import _securityContext
        from stuphos.kernel import Programmer

        # XXX billable?
        return _securityContext(Programmer(self._object.name))


MobileAdapter = MobileExtensionPoint


class ZoneResetAccess(writeprotected):
    '''
    A handle for providing restricted access to the zone object.
    It allows for the script to do anything that the zone can do.
    '''

    __public_members__ = ['initialized']

    @classmethod
    def _Get(self, zone):
        try: return zone.accessHandle
        except AttributeError:
            return self(zone) # Ephemeral.

            zone.accessHandle = a = self(zone)
            return a

    def __init__(self, zone):
        self.__zone = zone

    def __repr__(self):
        return '{zone #%d %r}' % (self.vnum, self.name)

    @property
    def initialized(self):
        return getattr(self.__zone, '_initialized', False)
    @initialized.setter
    def initialized(self, value):
        self.__zone._initialized = bool(value)

    @property
    def vnum(self):
        return self.__zone.vnum

    @property
    def name(self):
        return self.__zone.name


    @property
    def rooms(self):
        rooms = (RoomAdapter(r, perspective = self) for r in self.__zone.rooms)
        return vmCurrentTask().sequence(rooms)


    # @property
    # def mobiles(self):
    #     return dict((m.vnum, m) for m in map(self._MobilePrototypeAccess, self.__zone.mobiles))

    class _MobilePrototypeAccess(object):
        def __init__(self, mobile):
            self.__mobile = mobile

        @property
        def vnum(self):
            return self.__mobile.vnum

        @property
        def special(self):
            return self.__mobile.special

        @special.setter
        def special(self, procedure):
            # Todo: implement access-sensitive (secure) synchronous girl procedure.
            pass


ZoneSystemAccess = (ZoneResetAccess,)

class RoomExtensionPoint(verbal, perspective):
    _readonly_properties = []

    __public_members__ = ['interface', 'verbs']

    def __repr__(self):
        # Todo: repr(self.name) if self._perspectiveIs(MobileAdapter) and self._perspective._object.canSeeRoom(self._object) else '<darkness>'
        return f'<room$adapter {repr(self._object.name)}>'

    @property
    def vnum(self):
        self._perspectiveCheck(ZoneSystemAccess, 'Vnum not available!')
        return self._object.vnum


    @property
    def interface(self):
        # Todo: for ultimate PeerAdapter perspective, return constrainStructureMemory copy of interface.
        self._perspectiveCheck(ZoneSystemAccess, 'Interface not available!')
        return getattr(self._object, '_interface', None)

    @interface.setter
    def interface(self, value):
        self._perspectiveCheck(ZoneSystemAccess, 'Interface not available!')
        self._object._interface = value

    @property
    def verbs(self):
        # Todo: for ultimate PeerAdapter perspective, use verbCall API method to access.
        self._perspectiveCheck(ZoneSystemAccess, 'Verbs not available!')
        return getattr(self._object, '_verbs', None)

    @verbs.setter
    def verbs(self, value):
        self._perspectiveCheck(ZoneSystemAccess, 'Verbs not available!')
        self._object._verbs = value

RoomAdapter = RoomExtensionPoint



# #!/usr/local/bin/python
# from cherrypy import cpg

# class QueryPage:
# 	DEFAULT_LISTING_FORMAT = ''
# 	DEFAULT_QUERY_STRING = ''

# 	scanParams = { }

# 	@cpg.expose
# 	def index(self, **query):
# 		# Extract from web arguments programmable python filter logic and a listing
# 		# format for overriding the below scanWorld function.

# 		# Program the query into a scanner, and access the world.scanParts generator.
# 		def scanWorld(x):
# 			if self.DEFAULT_QUERY_STRING:
# 				return eval(self.DEFAULT_QUERY_STRING, self.__dict__)

# 			if self.DEFAULT_LISTING_FORMAT:
# 				if type(x) in (dict, str):
# 					return self.DEFAULT_LISTING_FORMAT % x
# 				else:
# 					return self.DEFAULT_LISTING_FORMAT % str(x)

# 			return str(x) + '<br>'


# 		# Page generation -- Apply scanner parameters.
# 		for x in self.pageControl.header():
# 			yield x
# 		for x in self.world.scanParts(**self.scanParams):
# 			yield scanWorld(x)
# 		for x in self.pageControl.footer():
# 			yield x

# 	shutdown = cpg.expose(lambda self:self.pageControl.stop())
# 	refresh  = cpg.expose(lambda self:self.pageControl.refresh())

# 	@cpg.expose
# 	def debug(self):
# 		from code import InteractiveConsole as IC
# 		IC(locals = globals()).interact(banner = None)

# 	def __init__(self, world, pageControl):
# 		self.world        = world
# 		self.pageControl  = pageControl

# 	def __repr__(self):
# 		# Recycle the index.
# 		# for x in self.index():
# 		#	yield x

# 		return ''.join(map(str, self.index()))

# from threading import Thread

# class PageLink(Thread):
# 	'Boots the page refresh server method.'
# 	startMap = {'serverPort':8080}

# 	def __init__(self, pagelink, startMap = {}):
# 		apply(super(PageLink, self).__init__)

# 		self.cpg = cpg
# 		self.pagelink = pagelink

# 		self.startMap = dict(self.startMap)
# 		self.startMap.update(startMap)

# 	def refresh(self):
# 		p = self.cpg.root = self.pagelink(self)
# 		return p

# 	def run(self):
# 		# Boot the web server interface.
# 		self.refresh()
# 		self.cpg.server.start(configMap = self.startMap)

# 	def start(self):
# 		apply(super(PageLink, self).start)
# 		return self

# 	def header(self):
# 		yield '<html>\n'
# 	def footer(self):
# 		yield '</html>\n'

# 	# Provide shutdown method
# 	def stop(self):
# 		self.cpg.server.stop()

# def bootQueryPage(world, async = False):
# 	def refresh(pageControl):
# 		web = QueryPage(world, pageControl)
# 		return web

# 	return apply(getattr(PageLink(refresh), async and 'start' or 'run'))

# DEFAULT_STUPHLIB_WORLD = ''

# if __name__ == '__main__':
# 	from os import environ
# 	from .wldlib import World

# 	bootQueryPage(World(environ.get('STUPH_PATH') or DEFAULT_STUPHLIB_WORLD))



# Command Messaging Base.
# --
# Copyright 2021 runphase.com .  All rights reserved
#
import email

from urllib.parse import urlparse
from urllib.parse import splitport

from traceback import print_exc

def breakOn(function):
    import pdb
    from ph import enqueueHeartbeatTask
    return lambda *args, **kwd:enqueueHeartbeatTask \
           (pdb.runcall, function, *args, **kwd)

X_CHECKPOINT_COMMAND_TYPE = 'X-Checkpoint-Command-Type'

def getMessageHeaders(msg):
    return dict(list(msg.items()))

def buildCommandFromMessage(message, headers = None):
    if headers is None:
        # This doesn't really have to exist anymore because of all headers
        # come from the message...
        headers = getMessageHeaders(message)

    cmd_type = headers.get(X_CHECKPOINT_COMMAND_TYPE)
    cmd_class = CommandClasses.get(cmd_type)
    if cmd_class is not None:
        return cmd_class.FromMessage(message, headers)

def isTrustedCommand(url, message):
    return message and message.CanExecuteFrom(url)

#@breakOn
def ExecuteCommandMessage(message):
    return message.Execute()

class CommandMessage:
    @classmethod
    def FromMessage(self, message, headers):
        payload = message.get_payload(decode = True) # the impl can handle encoding
        return self(payload, headers, message.get_content_type())

    def CanExecuteFrom(self, url):
        if self.__class__ is CommandMessage:
            return False # Must overload Execute.

        # !!! Security !!!
        # return True

        parsed = urlparse(url)
        if parsed.scheme.lower() == 'https':
            return True

        (hostname, port) = splitport(parsed.netloc)
        if hostname.lower() == 'localhost':
            return True # XXX This may not always be true!!

        return False # True if urlparse(url).scheme.lower() == 'https'

    def __init__(self, payload, headers = None, content_type = None):
        self.payload = payload
        self.headers = headers
        self.content_type = content_type

    def __repr__(self):
        content_type = self.getContentType()
        if content_type:
            content_type = ' (%s)' % content_type

        return '<%s%s>' % (self.__class__.__name__, content_type)

    def getContentType(self, default = None):
        return getattr(self, 'content_type', default)
    def getHeaders(self):
        return self.headers

    def __getitem__(self, name):
        return self.headers[name]
    def get(self, name, default):
        return self.headers.get(name, default)

class BatchCommandMessage(CommandMessage):
    def __init__(self, subcommands):
        self.subcommands = subcommands
    def __repr__(self):
        return '<%s: %d subcommands>' % (self.__class__.__name__,
                                         len(self.subcommands))

    @classmethod
    def FromMessage(self, message, headers):
        assert message.is_multipart()
        return self(list(map(buildCommandFromMessage, message.get_payload())))

    def Execute(self):
        for command in self.subcommands:
            if command is not None: # wtf, batches with their own default content.
                try: command.Execute()
                except:
                    # Todo: redirect this more appropriately:
                    print_exc()

# Registry.
CommandClasses = {}

def RegisterCommandClass(content_type, cmd_class):
    CommandClasses[content_type] = cmd_class

def AsCommandClass(content_type):
    # Decorator
    def register(object):
        RegisterCommandClass(content_type, object)
        return object

    return register

# Part of multiplexing commands.
RegisterCommandClass('application/x-stuph-command-batch', BatchCommandMessage)



from stuphos.kernel import Girl
import json, pdb

def tokenize(source):
	i = Girl.Module.prebuiltGrammar.tokenize

	for t in i(source):
		yield t

class Yield(Exception):
	def __init__(self, b):
		self.buffer = b

class Unget(Yield):
	def __init__(self, b, token):
		Yield.__init__(self, b)
		self.token = token

class Dedent(Yield):
	pass

def countws(n):
	x = 0

	for c in n:
		if c in '\r\t ':
			x += 1
		else:
			break

	return x

def listing(function):
	def call(*args, **kwd):
		return list(function(*args, **kwd))

	return call

joinNL = '\n'.join

def nling(function):
	def call(*args, **kwd):
		return joinNL(function(*args, **kwd))

	return call

@nling
def insertColons(source):
	lines = source.split('\n')
	previous = 0
	ln = None

	for n in lines:
		ws = countws(n)

		if ln is not None: # don't break on source-start whitespace.
			if ln[-1:] != ':' and (ln or ws > previous):
				yield ln + ':'
			else:
				yield ln

		previous = ws
		ln = n

	if ln is not None:
		yield ln


def parse(source):
	# XXX this is actually quirky with the tokenizer, because it's a single COLON,
	# so technically, all single-identifier token lines will always parse as TEXT blocks.
	#
	# The prevelant problem is that the indent lexer requires at least one COLON anyway
	# to produce indentation tokens, so it must be redesigned: either the lexer must
	# provide aCOLONic mode indentation, or something else.

	# Because this utilizes the gs_index_lexer tokenizer, it should parse inline TEXT tokens:
	# object method
	#	  arg
	#		  This text is passed as the first argument to object.method
	#		  and the TEXT token name (arg1) is ignored.
	#
	#	  arg
	#		  Similarly, this text string is passed as second argument.
	#		  The tokenizer requires an IDENTIFIER before COLON, but its
	#		  asymbolic.
	#

	# It could somehow recognize double colons, or insertColons could more intelligently
	# add colons so that a single colon works.

	# Another functionality could be something like this:
	# decorator
	# 	  function name$ (string arg1, string arg2)
	#	  	  return arg1 + arg2
	#
	# or
	# function name
	#	  code
	#	  	  def name(arg1, arg2):
	#		  	  return arg1 + arg2
	#
	# or
	# call
	# 	  attributeOf name
	# 	  	  module
	# 	  	  	  def name(arg1, arg2):
	# 		      	  return arg1 + arg2
	#	  string a
	#	  string b

	source = insertColons(source)
	source = source.replace('\t', '    ')
	return _parse(tokenize(source))

def _parse(i):
	b = []
	o = []
	e = ''

	# print '_parse'
	for token in i:
		# print token
		if token.type == 'ENDMARKER':
			# (Do not care if there is not any trailing newline.)
			if o:
				b.append(o) # technically this should be [o]

			return b

		if e == '':
			if token.type == 'NEWLINE':
				if o: # (blank lines)
					b.append([o])
					o = []

					# There can be no indent without a preceding header.
					e = 'indent'

			elif token.type == 'DEDENT':
				# print o

				if o: # ENDMARKER following this DEDENT
					b.append([o])

				raise Dedent(b)

			else:
				o.append(token)

		elif e == 'indent':
			if token.type == 'INDENT':
				# print 'subparse'

				# todo: quotational attribute grammatical lookahead.

				try: d = _parse(i)
				except Dedent as d:
					d = d.buffer

				# print d
				# print b[-1]
				b[-1].append(d)
				e = ''
				continue

			elif token.type == 'DEDENT':
				# really, this means terminate this _parse,
				# and the _parse before it.

				if o: # ENDMARKER following this DEDENT
					b.append([o])

				raise Dedent(b)

			elif token.type == 'NEWLINE':
				# XXX if o: b.append([o])!
				pass

			else:
				# XXX A newline doesn't NECESSITATE an indent.
				o.append(token)

				e = ''
				continue

	else:
		raise SyntaxError('Expected newline')


@nling
def dedent(source):
	lines = iter(source.split('\n'))

	for i in lines:
		n = countws(i)
		if len(i) != n:
			yield i[n:]
			break

	for i in lines:
		yield i[n:]

from ply.lex import LexToken
def newToken(type, value):
	x = LexToken()
	x.type = type
	x.value = value
	x.lineno = 0
	x.lexpos = 0
	return x

class extension:
	# Todo: implement this (don't rely on WRLC)
	pass

class Statement(list):
	def __init__(self, node, statements = []):
		self.node = node
		statements = list(statements)
		list.__init__(self, statements)

	@classmethod
	def List(self, statements):
		return self([], statements)

	@property
	def accessorize(self):
		last = None

		for n in self.node:
			if last is not None:
				if last.type == 'NAME':
					if n.type == 'NAME':
						yield newToken('DOT', '.')
						yield n

						last = n
						continue

					elif n.type == 'NUMBER':
						yield newToken('LBRACKET', '[')
						yield n
						yield newToken('RBRACKET', ']')

						last = n
						continue

				elif last.type == 'NUMBER':
					pass

			last = n
			yield n


	@property
	def nodeValues(self):
		return mapi(attributeOf.value, self.node)

	@property
	def accessorizedValues(self):
		return mapi(attributeOf.value, self.accessorize)

	@property
	def line(self):
		# Header line with WS.
		return ' '.join(mapi(str, self.nodeValues))

	@property
	def statements(self):
		return nls(mapi(str, self))

	@nling
	def __str__(self):
		yield self.line
		if self:
			yield indent(self.statements)
			yield ''

	@property
	def lineSource(self):
		return ''.join(mapi(str, self.accessorizedValues))

	@property
	def source(self):
		this = self.lineSource

		if self:
			if this:
				args = ', '.join(mapi(attributeOf.source, self))
				return '%s(%s)' % (self.lineSource, args)

			return '[%s]' % ', '.join(mapi(attributeOf.source, self))

		return this

	@property
	def html(self):
		attrs = ' '.join('%s=%r' % a for a in slices(self, 2))
		yield '<%s %s>' % (self.node[0].value, attrs)

		yield indent(mapi(attributeOf.html, self))
		yield '</%s>' % self.node[0].value

	@property
	def python(self):
		return self.Python(self)

	class Python(extension):
		@property
		def compiled(self):
			pass

	@property
	def emulated(self):
		return self.Emulated(self)

	class Emulated(extension):
		@property
		def compiled(self):
			from stuphos.kernel import Girl
			return Girl(Girl.Expression, self.source)

		@runtime.available(runtime.System.Engine)
		def evaluating(vm, self, **environ):
			from stuphos.kernel import Script, findCurrentUser
			procedure = self.compiled

			with vm.Loading(Script, environ = environ, user = findCurrentUser()) as task:
				procedure.setEnvironment(task.environ) # Unit
				task.addFrameCall(procedure)
				runtime[runtime.System.Journal].waitLogs \
					(task, lambda log, task, traceback:
					 system.print_(nls('%s %s\n%s\n' % \
					 	(f.procedure[i-1], i, indent(nls(mapi(repr, task.stack))))
					 	for (f, i) in traceback)))

				return task


def _syntaxOf(i):
	# print i
	for x in i:
		# print x
		if isinstance(x[0], list): # a line + statements node
			if len(x) > 1: # full
				yield Statement(x[0][:-1], _syntaxOf(x[1])) # -COLON
			else: # a non-dedented, non-parsed statements line node
				# (or, a node with no statements..)
				yield Statement(x[0][:-1])

		else: # a line (- statements) node
			yield Statement(x)

syntaxOf = listing(_syntaxOf)

def compile(source):
	return Statement.List(_syntaxOf(parse(source)))


def shortateCall(frame, source, locals = None):
	'''
	frame task frameCall
		attributeOf compiled
			compile
				source

		locals
	'''

	procedure = compile(source).emulated.compiled

	# def sendToOperator(line, multiline = False, backlog = False):
	# 	print line

	# frame.task.tracing = frame.task.auditInstruction
	# frame.task.sendToOperator = sendToOperator

	# import pdb; pdb.set_trace()

	return frame.task.frameCallSwap(procedure, locals = locals)


def shortate(name, value, **kwd):
	'''
	set node
		attributeOf compiled
			compile
				value

	wrap($class):
		base: extension

		do($method):
			parameters: [*args]
			code::
				pass

	function($shortate)::
		this do
			these are more
			things to do

	run($trigger)::
		this = environment.wrap($('service/wrap', this))
		environment.function(mapping(['this', this]))

	'''

	node = compile(value).emulated.compiled

	def shortateCall(frame, locals = None):
		frame.task.frameCall(node, locals = locals)

	return native

def shortize(function):
	'''
	@shortize
	def lookup(name):
		"""
		accessorOf loading structure value code
			getAttribute my
				name

		"""

	# lookup('jobs').outcome

	'''

	s = compile(dedent(docOf(function))).emulated

	def call(*args, **kwd):
		# kwd.update(dict(zip(function.func_code.co_argnames, args)))
		return s.evaluating(**kwd)

	return call


def parsex():
	# Compile parsed documentation source.
	#  (/ph/lang/layer/gs_indent_lexer.py:139) raise IndentationError("indentation increase but not in new block")
	'''
	this is a line:
		this is a subline:
			subline 1
			subline 2

		subline 3
		subline 4:
			x

	another line 1:
		subline 5

	another line 2

	a:
		b:
			c:
				d:
					e:
						f
							g
	'''

	def f(i):
		if isinstance(i, (list, system.types.generator)):
			return list(map(f, i))

		return i.value

	concating = lambda f: lambda *args, **kwd: ''.join(f(*args, **kwd))

	#@concating
	@nling
	def p(i):
		for x in i:
			if len(x) > 1 and isinstance(x[0], list):
				if isinstance(x[0][0], list):
					yield indent(p(x[0]))
				else:
					yield ' '.join(mapi(str, x[0]))
					yield indent(p(x[1]))

				if len(x) > 2:
					yield str(x[2:])

			elif isinstance(x[0], list):
				yield ' '.join(mapi(str, x[0]))
			else:
				yield ' '.join(mapi(str, x))


	source = docOf(parsex)
	source = dedent(source)
	# print source

	return compile(source)

	# return eat(tokenize(source))
	result = f(parse(source))
	return p(result)

	return json.dumps(result, indent = 1)



# UUID Naming Hierarchy -- to be merged with runtime registry.
# But note, that these are two different concepts.
from uuid import uuid5 as CreateUUIDFrom, NAMESPACE_DNS

def CreateUUIDFromUrl(url, root = NAMESPACE_DNS):
    return CreateUUIDFrom(root, url)

class NamingStructure:
    def __init__(self, root, name):
        self.root = root
        self.name = name

    def __getitem__(self, subname):
        return NamingStructure(CreateUUIDFrom(self.root, subname), subname)

    def getID(self):
        return self.root
    def getName(self):
        return self.name

# Some roots.
PENOBSCOTROBOTICS = NamingStructure(CreateUUIDFromUrl('penobscotrobotics.us'), 'Penobscot Robotics')

STUPHMUD = PENOBSCOTROBOTICS['stuphmud.net']
STUPHOS = STUPHMUD['stuphos.penobscotrobotics.us']

# Todo: runtime api integration

# @runtime.api.System.Naming
# @apply
# class API:
#     PENOBSCOTROBOTICS = PENOBSCOTROBOTICS
#     STUPHMUD = STUPHMUD
#     STUPHOS = STUPHOS
